// ENCODING: UTF-8
'use strict';

// DEPENDS ON GLOBALS:
// - settings

// START EXPLORER CLASS
module.exports = (function () {

// HERE PRIVATE STATIC VARS

// BUILT-IN MODULES
var cp = require('child_process');
var ev = require('events');
var fs = require('fs');
var qs = require('querystring');
var ut = require('util');

// EXTERNAL MODULES
var fx = require('./lib/fx-utils');
var qt = require('./lib/node-qt');

var File = require('./lib/file');


/**
 * Explorer Object
 * @constructor
 */
var Explorer = function(parent){
	this._setup(parent);
}
ut.inherits(Explorer, ev.EventEmitter);


//################################################
// PRIVATE METHODS
//################################################

/**
 * init
 *
 * @this {Explorer}
 */
Explorer.prototype._setup = function(parent){
	var self = this;

	// INSTANCE VARS
	self.hsplitter = new qt.QSplitter(parent); // container
	self.treeView = new qt.QTreeView(parent, true);
	self.tablePanel = new qt.QWidget(parent);
	self.tableComboBox = new qt.QComboBox(parent);
	self.tableView = new qt.QTableView(parent, true);

	self.treeViewFileModel = new qt.QFileSystemModel;
	self.tableViewFileModel = new qt.QFileSystemModel;

	self.connected = false;
	self.path = undefined; // always ends on '/'
	self.actions = {};

	self._setupSplitter();
	self._setupTreeView();
	self._setupTableView();
	self._setupComboBox();
}

/**
 * _setupSplitter
 *
 * @this {Explorer}
 */
Explorer.prototype._setupSplitter = function(){
	var self = this;

	self.hsplitter.setOrientation(1);
	self.hsplitter.addWidget(self.treeView);

	var vl = new qt.QVBoxLayout();
	vl.setSpacing(5);
	vl.setContentsMargins(0,0,0,0);
	vl.addWidget(self.tableComboBox );
	vl.addWidget(self.tableView);
	self.tablePanel.setLayout(vl);
	self.hsplitter.addWidget(self.tablePanel);

	self.hsplitter.setCollapsible(0, false);
	self.hsplitter.setCollapsible(1, false);
	self.hsplitter.setSizes([150,250]);
}

/**
 * _setupTreeView
 *
 * @this {Explorer}
 */
Explorer.prototype._setupTreeView = function(){
	var self = this;
	var currentIndex;
	var currentDragIndex;

	self.treeView.setHeaderHidden(true);
	self.treeView.setAcceptDrops(true);
	self.treeViewFileModel.setFilter(qt.QDIR.ALLDIRS | qt.QDIR.NODOTANDDOTDOT);

	// CREATE CONTEXT MENU

	self.treeView.contextMenu = new qt.QMenu(self.treeView);

	self.actions.treeSubfolder = this.treeView.contextMenu.addAction('New Subfolder...').on('triggered', function(){
		self._maybeMakeDir();
	});

	// EVENT HANDLERS

	self.treeView.on('contextMenuEvent', function(e){
		var modelIndex = self.treeView.indexAt(e.pos());
		if (modelIndex.isValid()){
			self.treeView.contextMenu.exec(e.globalPos());
		}
	});

	self.treeView.on('dragEnterEvent', function(p){
		currentIndex = self.treeView.currentIndex();
		currentDragIndex = self.treeView.indexAt(p);
	});

	self.treeView.on('dragMoveEvent', function(p){
		var index = self.treeView.indexAt(p);
		if (!index.equals(currentDragIndex)){
			self.treeView.setSelectedIndexes([currentIndex, index], qt.SELECTIONFLAG.CLEAR | qt.SELECTIONFLAG.SELECT);
			currentDragIndex = index;
		}
	});

	self.treeView.on('dragLeaveEvent', function(p){
		if (currentIndex){
			self.treeView.setSelectedIndexes([currentIndex], qt.SELECTIONFLAG.CLEAR | qt.SELECTIONFLAG.SELECT);
			currentIndex = false;
		}
	});

	self.treeView.on('dropEvent', function(e){
		if (currentIndex){
			self.treeView.setSelectedIndexes([currentIndex], qt.SELECTIONFLAG.CLEAR | qt.SELECTIONFLAG.SELECT);
			currentIndex = false;
		}
		if (!self.connected) return; // e.ignore();

		// GET LOCAL TARGET PATH
		var modelIndex = self.treeView.indexAt(e.pos());
		if (modelIndex==undefined) return;
		var path = self.treeViewFileModel.filePath(modelIndex);
		if (path.charAt(path.length-1)!='/') path+= '/';
		var src = e.source();

		if (src=='explorer.tableView'){

			//1. FROM LOCAL TABLE

			var data = e.mimeData(true)['text/uri-list'];
			if (data==undefined) return;
			var items = data.split('\r\n');
			items.pop();

			// PARSE
			var files = [];
			items.forEach(function(f){
				if (f.substr(0,8)=='file:///') f = f.substr(8);
				f = qs.unescape(f);
				var name = f.split('/').pop();
				fs.rename(f, path+name);// renameSync NOW MOVE the files/folders to path!
			});

		}else if (src=='ftpExplorer.tableWidget'){

			//2. FROM REMOTE TABLE

			var data = e.mimeData(true)['text/table-cells']; //text/plain
			if (data==undefined) return;
			var cells = data.split('\r\n');
			cells.pop();

			var lineCnt = cells.length/3;
			var itemNames = [];
			for (var i=0;i<lineCnt;i++){
				itemNames.push(cells[i]);
			}

			// HERE CONFIRM DIALOG?
			self.emit('downloadRequested', itemNames, path); // BOTH FILES AND FOLDERS!

		}else{

			// 3. ALSO HANDLE DROPS FROM EXPLORER !?
			// IF YES, COPY OR MOVE? ASK VIA DIALOG?

		}
	});
}

/**
 * _setupTableView
 *
 * @this {Explorer}
 */
Explorer.prototype._setupTableView = function(){
	var self = this;

	var filter = qt.QDIR.DIRS | qt.QDIR.FILES | qt.QDIR.NODOT;
	if (settings.showHiddenFiles) filter |= qt.QDIR.HIDDEN;
	self.tableViewFileModel.setFilter(filter);

	self.tableView.setObjectName('explorer.tableView');
	self.tableView.setDragEnabled(true);
	self.tableView.setShowGrid(settings.showGrid);
	self.tableView.setAlternatingRowColors(settings.alternatingRowColors);
	self.tableView.setSortingEnabled(true);
	self.tableView.setRowDefaultHeight(17);
	self.tableView.setVerticalHeaderVisible(false);
	self.tableView.setHorizontalStretchLastSection(true);
	self.tableView.setSelectionBehavior(qt.SELECTIONBEHAVIOR.SELECTROWS);

//	self.tableView.setColumnWidth(0,120);
//	self.tableView.setColumnWidth(1,60);
//	self.tableView.setColumnWidth(2,60);
//	self.tableView.setColumnWidth(3,120);

	self.tableView.setEditTriggers(0); // make non-editable

	self.tableView.setStyleSheet("QHeaderView::section{padding-top:3px;padding-bottom:2px;padding-left:0};");

	// CREATE CONTEXT MENU

	self.tableView.contextMenu = new qt.QMenu(self.tableView);

	self.actions.upload = self.tableView.contextMenu.addAction('Upload selected');
	self.actions.upload.on('triggered', function(){
		self._uploadSelection();
	});
	
	self.actions.rename = self.tableView.contextMenu.addAction('Rename selected');
	self.actions.rename.on('triggered', function(){
		self._maybeRenameSelectedItem();
	});
	
	self.actions.delete = self.tableView.contextMenu.addAction('Delete selected');
	self.actions.delete.on('triggered', function(){
		self._maybeDeleteSelectedItems();
	});

	self.tableView.contextMenu.addSeparator();

	self.actions.open = self.tableView.contextMenu.addAction('Open selected');
	self.actions.open.on('triggered', function(){
		var indexes = self.tableView.selectedIndexes();
		if (indexes.length==0) return;
		var file = self.path + indexes[0].data();
		cp.exec('start "" "'+file+'"');
	});
	
	self.tableView.contextMenu.addSeparator();
	
	self.tableView.contextMenu.addAction('Refresh').on('triggered', function(){
		self.update();
	});
	
	// EVENT HANDLERS
	self.tableView.on('contextMenuEvent', function(e){
		if (!self.connected) return;
		var modelIndex = self.tableView.indexAt(e.pos());
		var isValid = modelIndex.isValid();
		var isDir = isValid && self.tableViewFileModel.isDir(modelIndex);
		
		//if (isValid){
		self.actions.upload.setEnabled(isValid);
		self.actions.rename.setEnabled(isValid);
		self.actions.delete.setEnabled(isValid);
		self.actions.open.setEnabled(isValid && !isDir);
			
		self.tableView.contextMenu.exec(e.globalPos());
		//}
	});

	self.tableView.on('doubleClicked', function(modelIndex){
		if (self.tableViewFileModel.isDir(modelIndex)){
			self.path = self.tableViewFileModel.filePath(modelIndex);
			self.treeView.setCurrentIndex(self.treeViewFileModel.setRootPath(self.path));
			self.tableView.setRootIndex(self.tableViewFileModel.setRootPath(self.path));
		}else{

			// A: OPEN
			var file = self.path + modelIndex.data();
			cp.exec('start "" "'+file+'"');

			// B: UPLOAD
			//self._uploadSelection([modelIndex]);
		}
	});

	self.tableView.on('keyPressEvent', function(e){
		if (e.key()==qt.KEY.KEY_DELETE){
			self._maybeDeleteSelectedItems();
		}
	});

	self.tableView.on('dragMoveEvent', function(p){
		var modelIndex = self.tableView.indexAt(p);
		if (modelIndex.column()==0 && self.tableViewFileModel.isDir(modelIndex)){
			self.tableView.setSelectedIndexes([modelIndex], qt.SELECTIONFLAG.CLEAR | qt.SELECTIONFLAG.SELECT | qt.SELECTIONFLAG.ROWS); //qt.SELECTIONFLAG.CLEAR | qt.SELECTIONFLAG.SELECT
		}else{
			self.tableView.setSelectedIndexes([], qt.SELECTIONFLAG.CLEAR);
		}
	});


	// WE ONLY KNOW FILENAMES, NOT THE REMOTE PATH
	self.tableView.on('dropEvent', function(e){
		if (!self.connected) return;
		var src = e.source();

		if (src=='ftpExplorer.tableWidget') {

			// A) FILES DRAGGED FROM FTPEXPLORER TABLE
			var data = e.mimeData(true)['text/table-cells']; //text/plain
			if (data==undefined) return;
			var cells = data.split('\r\n');
			cells.pop();
			var lineCnt = cells.length/3;
			var itemNames = [];
			for (var i=0;i<lineCnt;i++){
				itemNames.push(cells[i]);
			}
			self.emit('downloadRequested', itemNames); // BOTH FILES AND FOLDERS!

		}else if (src=='explorer.tableView'){

			// B) FILES DRAGGED WITHIN EXPLORER TABLE
			var modelIndex = self.tableView.indexAt(e.pos());
			if (self.tableViewFileModel.isDir(modelIndex)){
				var path = self.tableViewFileModel.filePath(modelIndex);
				if (path.charAt(path.length-1)!='/') path+= '/';
				var data = e.mimeData(true)['text/uri-list'];
				if (data==undefined) return;
				var items = data.split('\r\n');
				items.pop();
				var files = [];
				items.forEach(function(f){
					if (f.substr(0,8)=='file:///') f = f.substr(8);
					f = qs.unescape(f);
					var name = f.split('/').pop();
					fs.rename(f, path+name);// renameSync NOW MOVE the files/folders to path!
				});
			}
		}
	});
}

/**
 * _setupComboBox
 *
 * @this {Explorer}
 */
Explorer.prototype._setupComboBox = function(){
	var self = this;

	self.tableComboBox.on('activated', function(index){
		self.path = self.tableComboBox.itemText(index);
		self.tableViewFileModel.setRootPath(self.path);
		self.treeView.setCurrentIndex(self.treeViewFileModel.setRootPath(self.path));
		self.tableView.setRootIndex(self.tableViewFileModel.setRootPath(self.path));
	});
}

/**
 * Downloads currently selected (or passed as QModelIndex list) files and folders in current dir
 *
 * @param {string} indexes - Array of indexes. [optional]
 */
Explorer.prototype._uploadSelection = function(indexes){
	var self = this;

	if (indexes==undefined) indexes = self.tableView.selectedIndexes();
	if (indexes.length==0) return;

	var files = []; // [ File(name, path, size, date) ]

	// selection can be list of files and folders (flat, i.e. only in current path)
	indexes.forEach(function(index){
		if (index.column()==0){
			var name = index.data(); //index.filePath();
			if (name=='..') return;
			var path = self.path + name; //index.filePath();
			var stats = fs.statSync(path);
			if (stats.isFile()){
				files.push( new File(name, path, stats.size) );
			}else if (stats.isDirectory()){
				fx.readDir(path, files);
			}
		}
	});

	self.emit('uploadRequested', files); // ONLY FILES, NO FOLDERS!
}

/**
 * Asks user for new name, if result is valid, renames the selected file/folder
 *
 */
Explorer.prototype._maybeRenameSelectedItem = function(){
	var self = this;

	var indexes = self.tableView.selectedIndexes(); // -> array of QModelIndex: column,row,data
	if (indexes.length==0) return;
	var oldName = indexes[0].data();
	var path = self.path + oldName;
	var stats = fs.statSync(path);
	var isFolder = stats.isDirectory();
	var name = qt.QInputDialog.getText('Rename '+(isFolder?'folder':'file'),
	    'Enter a new name:', oldName);
	if (typeof name!='string' || name=='' || name==oldName) return;
	if (settings.illegalChars){
		name = name.replace(new RegExp('['+settings.illegalChars+']', 'g'), '');
	}
	fs.rename(path, self.path+name); // was: renameSync
}

/**
 * Asks for confirmation, then deletes (recycles) currently selected files/folders
 *
 */
Explorer.prototype._maybeDeleteSelectedItems = function(){
	var self = this;

	var indexes = self.tableView.selectedIndexes(); // ->array of QModelIndex: column,row,data
	if (indexes.length==0) return;

	if (settings.showConfirmDialog){
		var ok = indexes.length==3?
			qt.confirm('Delete selected item?') :
			qt.confirm('Delete selected items?'); // files/folders
		if (!ok) return;
	}

	if (settings.useRecycleBin){
		var args = ['-f'];
		indexes.forEach(function(index){
			if (index.column()==0){
				var name = index.data();
				if (name=='..') return;
				args.push(self.path+name);
			}
		});
		cp.execFile('bin/recycle.exe', args, {}, function(err){
			if (err) return self.emit('error', err);
		});
	}else{
		indexes.forEach(function(index){
			if (index.column()==0){
				var name = index.data();
				if (name=='..') return;
				var path = self.path + name;
				var stats = fs.statSync(path);
				if (stats.isDirectory()){
					fx.deleteFolder(path);
				}else{
					fs.unlinkSync(path);
				}
			}
		});
	}
}

/**
 * Asks user for new folder name, if result is valid, creates new folder
 *
 */
Explorer.prototype._maybeMakeDir = function(){
	var self = this;

	var name = qt.QInputDialog.getText('Create new folder', 'Enter the name:', '');
	if (typeof name!='string' || name=='') return;

	if (settings.illegalChars){
		name = name.replace(new RegExp('['+settings.illegalChars+']', 'g'), '');
	}

	fs.mkdirSync(this.path+name);
}


//################################################
// PUBLIC METHODS
//################################################

Explorer.prototype.loadFilesystem = function(){
	var self = this;

	self.treeViewFileModel.setRootPath('My Computer');
	self.tableViewFileModel.setRootPath('My Computer');

//	var i = self.treeViewFileModel.index(dir);
//	self.treeView.setCurrentIndex(i);

	self.treeView.setModel(self.treeViewFileModel);
	self.tableView.setModel(self.tableViewFileModel);

	self.treeView.setColumnHidden(1, true);
	self.treeView.setColumnHidden(2, true);
	self.treeView.setColumnHidden(3, true);

	self.treeView.on('currentRowChanged', function(newModelIndex){
		self.path = self.treeViewFileModel.filePath(newModelIndex);

		var pos = self.tableComboBox.findText(self.path, 0);
		if (pos<0){
			self.tableComboBox.addItem(self.path);
			pos = self.tableComboBox.count()-1;
		}
		self.tableComboBox.setCurrentIndex(pos);

		if (self.path.charAt(self.path.length-1)!='/') self.path+= '/';
		self.tableView.setRootIndex(self.tableViewFileModel.setRootPath(self.path) );
	});

	// TEST: LOAD DEFAULT DIR
	var dir = __filename.split('\\').shift();
	self.path = dir;

	self.tableView.setAcceptDrops(true);
	self.connected = true;
	
	var i = self.treeViewFileModel.index(dir);
	self.treeView.setCurrentIndex(i);

	// self.treeView.sortByColumn(0,0);
}

Explorer.prototype.getCurrentPath = function(){
	var self = this;

	return self.path;
}

/**
 * Forces update of current view in table
 */
Explorer.prototype.update = function(){
	var self = this;

	// force explorer update by creating a new filemodel object
	var fileModel = new qt.QFileSystemModel;
	fileModel.setRootPath(self.path);
	
	var filter = qt.QDIR.DIRS | qt.QDIR.FILES | qt.QDIR.NODOT;
	if (settings.showHiddenFiles) filter |= qt.QDIR.HIDDEN;
	fileModel.setFilter(filter);
	
	var index = fileModel.setRootPath(self.path);
	self.tableView.setModel(fileModel);
	self.tableView.setRootIndex(index);

	delete self.tableViewFileModel;
	self.tableViewFileModel = fileModel;
}

/**
 * Returns the container widget (a QSplitter in this case)
 *
 * @this {Explorer}
 */
Explorer.prototype.container = function(){
	var self = this;

	return self.hsplitter;
}

return Explorer;
})();// END CLASS

