// ENCODING: UTF-8
'use strict';

// DEPENDS ON GLOBALS:
// - settings
// - sites

// START FTPEXPLORER CLASS
module.exports = (function () {

// HERE PRIVATE STATIC VARS

// BUILT-IN MODULES
var cp = require('child_process');
var ev = require('events');
var fs = require('fs');
var qs = require('querystring');
var ut = require('util');

// EXTERNAL MODULES
var async = require('./lib/async');
var fx = require('./lib/fx-utils');
var qt = require('./lib/node-qt');

var File = require('./lib/file');
var FtpClient = require('./lib/ftp');


/**
 * FTPExplorer Object
 * @constructor
 */
var FTPExplorer = function(parent){
	this._setup(parent);
}
ut.inherits(FTPExplorer, ev.EventEmitter);


//################################################
// PRIVATE METHODS
//################################################

/**
 * _setup
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._setup = function(parent){
	var self = this;

	// INSTANCE VARS
	self.hsplitter = new qt.QSplitter(parent); // container
	self.treeWidget = new qt.QTreeWidget(parent, true);
	self.tablePanel = new qt.QWidget(parent);
	self.tableComboBox = new qt.QComboBox(parent);
	self.tableWidget = new qt.QTableWidget(parent, true);

	self.client = new FtpClient();

	self.connected = false;
	self.currentSite = undefined;
	self.actions = {};
	self.rootItem = new qt.QTreeWidgetItem;
	self.currentTreeItem = self.rootItem;

	self.encoding = 'binary'; // ???

	//test
	self._debug = false;
	//self.streams = {};

	self.path = undefined;

	self._setupSplitter();
	self._setupTreeWidget();
	self._setupTableWidget();
	self._setupComboBox();
	self._setupClient();
}

/**
 * _setupSplitter
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._setupSplitter = function(){
	var self = this;

	self.hsplitter.setOrientation(1);

	self.hsplitter.addWidget(this.treeWidget);

	var vl = new qt.QVBoxLayout();
	vl.setSpacing(5);
	vl.setContentsMargins(0,0,0,0);
	vl.addWidget(self.tableComboBox );
	vl.addWidget(self.tableWidget);
	self.tablePanel.setLayout(vl);
	self.hsplitter.addWidget(self.tablePanel);

	//self.hsplitter.addWidget(this.tableWidget);

	self.hsplitter.setCollapsible(0, false);
	self.hsplitter.setCollapsible(1, false);

	self.hsplitter.setSizes([150,250]);
}

/**
 * _setupTreeWidget
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._setupTreeWidget = function(){
	var self = this;
	var currentIndex;

	self.treeWidget.setHeaderHidden(true);
	self.rootItem.setText(0, '/');
	self.rootItem.setFlags(qt.ITEMFLAG.ITEMISENABLED|qt.ITEMFLAG.ITEMISSELECTABLE);
	//rootItem.setIcon(0, __dirname + '/../assets/icons/root.png');
	self.treeWidget.addTopLevelItem(this.rootItem);

	self.treeWidget.setDragDropMode(qt.DRAGDROPMODE.DRAGDROP);
	self.treeWidget.setAcceptDrops(true);
	//self.treeWidget.setDragEnabled(true);

	//self.treeWidget.setStyleSheet("QTreeWidget {show-decoration-selected:0;} QTreeWidget::item:selected {background-color:#ffbbbb;}");

	// CREATE CONTEXT MENU
	self.treeWidget.contextMenu = new qt.QMenu(self.treeWidget);

	self.treeWidget.contextMenu.addAction('New Subfolder...').on('triggered', function(){
		self._maybeMakeDir();
	});


	// EVENT HANDLERS
	self.treeWidget.on('contextMenuEvent', function(e){
		if (!self.connected) return;
		var treeItem = self.treeWidget.itemAt(e.pos());
		if (treeItem){
			self.treeWidget.contextMenu.exec(e.globalPos());
		}
	});

	self.treeWidget.on('currentItemChanged', function(newItem){
		self.currentTreeItem = newItem;

		//		var itemPathList = self._getPathToCurrentTreeItem();
		//		itemPathList.shift(); // remove rootItem ('/')
		//		var path = '/';
		//		itemPathList.forEach(function(item){
		//			path += item.text()+'/';
		//		});

		var path = self._getPathToCurrentTreeItem();

		// first cache lookup?
		self.client.cwd(path, function(err, currentDir){
			if (err) return self.emit('error', err);
			self.path = path;
			self.update();

			// change to existing or add new path to comboBox
			var pos = self.tableComboBox.findText(self.path, 0);
			if (pos<0){
				self.tableComboBox.addItem(self.path);
				pos = self.tableComboBox.count()-1;
			}
			self.tableComboBox.setCurrentIndex(pos);
		});
	});

	self.treeWidget.on('dragEnterEvent', function(p){
		currentIndex = self.treeWidget.currentIndex();
	});

	self.treeWidget.on('dragMoveEvent', function(p){
		self.treeWidget.setSelectedIndexes([self.treeWidget.indexAt(p)], qt.SELECTIONFLAG.CLEAR | qt.SELECTIONFLAG.SELECT);
		//self.treeWidget.setCurrentIndex(self.treeView.indexAt(p));
	});

	self.treeWidget.on('dragLeaveEvent', function(p){
		if (currentIndex){
			self.treeWidget.setSelectedIndexes([currentIndex], qt.SELECTIONFLAG.CLEAR | qt.SELECTIONFLAG.SELECT);
			currentIndex = false;
		}
	});

	self.treeWidget.on('dropEvent', function(e){
		if (currentIndex){
			self.treeWidget.setSelectedIndexes([currentIndex], qt.SELECTIONFLAG.CLEAR | qt.SELECTIONFLAG.SELECT);
			currentIndex = false;
		}
		if (!self.connected) return; // e.ignore();

		// GET TARGET TREEITEM
		var treeItem = self.treeWidget.itemAt(e.pos());
		if (treeItem==undefined || treeItem.equals(self.currentTreeItem)) return;

		var targetPath = self._getPathToTreeItem(treeItem);
		var src = e.source();

		if (src=='ftpExplorer.tableWidget') {

			//1. FROM REMOTE TABLE

			//if (treeItem.equals(self.currentTreeItem)) return;
			if (targetPath==self.path) return;

			var data = e.mimeData(true)['text/table-cells']; //text/plain
			if (data==undefined) return;

			self._maybeMoveSelectedItems(targetPath);


		}else { //if (src=='explorer.tableView')

			//2. FROM LOCAL TABLE OR FROM EXPLORER!

			var data = e.mimeData(true)['text/uri-list'];
			if (data==undefined) return;
			var items = data.split('\r\n');
			items.pop();

			// PARSE
			var files = [];
			items.forEach(function(f){
				if (f.substr(0,8)=='file:///') f = f.substr(8);
				f = qs.unescape(f);
				var stats = fs.statSync(f);
				if (stats.isFile()){
					////new File(name, path, size)
					var name = f.split('/').pop();
					files.push( new File(name, f, stats.size) );
				}else if (stats.isDirectory()){
					fx.readDir(f, files);
				}
			});

			self.emit('uploadRequested', files,   targetPath);

		}

//			// 3. ALSO HANDLE DROPS FROM EXPLORER !?
//
//			var data = e.mimeData(true)['text/uri-list'];
//			if (data==undefined) return;
//			var items = data.split('\r\n');
//			items.pop();
//
//		}

	});
}

/**
 * _setupTableWidget
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._setupTableWidget = function(){
	var self = this;

	self.tableWidget.setObjectName('ftpExplorer.tableWidget');
	self.tableWidget.setShowGrid(settings.showGrid);
	self.tableWidget.setAlternatingRowColors(settings.alternatingRowColors);
	self.tableWidget.setSortingEnabled(true);
	self.tableWidget.setRowDefaultHeight(17);
	self.tableWidget.setColumnCount(4);
	self.tableWidget.setVerticalHeaderVisible(false);
	self.tableWidget.setHorizontalHeaderLabels(['Name','Size','Type','Date Modified']);
	self.tableWidget.setHorizontalStretchLastSection(true);
	self.tableWidget.setHorizontalHeaderDefaultAlignment(1);
	self.tableWidget.setSelectionBehavior(1); // QAbstractItemView::SelectRows

	self.tableWidget.setColumnWidth(0,120);
	self.tableWidget.setColumnWidth(1,60);
	self.tableWidget.setColumnWidth(2,60);
	self.tableWidget.setColumnWidth(3,60);

	self.tableWidget.setEditTriggers(0); // make non-editable

	self.tableWidget.setDragDropMode(qt.DRAGDROPMODE.DRAGDROP);

	// CREATE CONTEXT MENU

	self.tableWidget.contextMenu = new qt.QMenu(self.tableWidget);

	self.actions.download = self.tableWidget.contextMenu.addAction('Download selected');
	self.actions.download.on('triggered', function(){
		self.downloadSelection();
	});

	self.actions.rename = self.tableWidget.contextMenu.addAction('Rename selected');
	self.actions.rename.on('triggered', function(){
		self._maybeRenameSelectedItem();
	});

	self.actions.delete = self.tableWidget.contextMenu.addAction('Delete selected');
	self.actions.delete.on('triggered', function(){
		self._maybeDeleteSelectedItems();
	});

	self.tableWidget.contextMenu.addSeparator();

	self.actions.open = self.tableWidget.contextMenu.addAction('Open selected');
	self.actions.open.on('triggered', function(){//(tmp. Download)
		var items = self.tableWidget.selectedItems();
		if (items.length==0) return;
		var item = items[0];
		self.previewFile(item.text());
	});

	self.tableWidget.contextMenu.addSeparator();

	self.tableWidget.contextMenu.addAction('Refresh').on('triggered', function(){
		self.update();
	});

	// EVENT HANDLERS

	self.tableWidget.on('contextMenuEvent', function(e){
		if (!self.connected) return;

		var tableItem = self.tableWidget.itemAt(e.pos());
		//if (tableItem){

		self.actions.download.setEnabled(tableItem);
		self.actions.rename.setEnabled(tableItem);
		self.actions.delete.setEnabled(tableItem);
		self.actions.open.setEnabled(tableItem && !tableItem.isFolder());

		self.tableWidget.contextMenu.exec(e.globalPos());
		//}

	});

	self.tableWidget.on('cellDoubleClicked', function(row, col){
		var item = self.tableWidget.item(row, 0);
		var newTreeItem;
		if (item==undefined) return;
		var name = item.text();

		if (name=='..' || item.isFolder()){
			if (name=='..'){
				//if (self.path=='/') return; // ignore if already top!
				if (self.currentTreeItem.equals(self.rootItem)) return;
				newTreeItem = self.currentTreeItem.parent();
			}else{
				// find associated treeWidgetItem
				var cnt = self.currentTreeItem.childCount();
				for (var i=0;i<cnt;i++){
					if (self.currentTreeItem.child(i).text()==name){
						newTreeItem = self.currentTreeItem.child(i);
						break;
					}
				}
			}
			self.treeWidget.setCurrentItem(newTreeItem);

		}else{ // FILE

			// A: DOWNLOAD FILE
			//self.downloadSelection([name]);

			// B: DOWNLOAD TO TMP AND OPEN
			self.previewFile(name);
		}
	});

	self.tableWidget.on('keyPressEvent', function(e){
		if (e.key()==qt.KEY.KEY_DELETE){
			self._maybeDeleteSelectedItems();
		}
	});

	self.tableWidget.on('dragMoveEvent', function(p){
		var tableItem = self.tableWidget.itemAt(p);
		if (tableItem && tableItem.isFolder())
			self.tableWidget.setCurrentItem(tableItem, qt.SELECTIONFLAG.CLEAR | qt.SELECTIONFLAG.SELECT | qt.SELECTIONFLAG.ROWS);
		else
			self.tableWidget.setCurrentCell(-1,-1);

		//else
		//	self.tableWidget.setCurrentItem(0);

		//var modelIndex = self.tableView.indexAr(p);

//		if (self.tableViewFileModel.isDir(modelIndex)){
//			self.tableView.setCurrentIndex(modelIndex);
//		}else{
//			self.tableView.setSelectedIndexes([], 1 | 2);
//		}
	});

	self.tableWidget.on('dropEvent', function(e){
		if (!self.connected) return; // e.ignore();
		var src = e.source();

		if (src=='explorer.tableView'){

			// FILES DRAGGED FROM LOCAL EXPLORER TABLE
			var data = e.mimeData(true)['text/uri-list'];
			if (data==undefined) return;
			//{ 'text/uri-list': 'file:///D:/DATA/MP3Diags_trace.txt\r\n' }
			//{ 'text/uri-list': 'file:///C:/Neuer%20Ordner\r\n' }

			var items = data.split('\r\n');
			items.pop();

			// PARSE
			var files = [];
			items.forEach(function(f){
				if (f.substr(0,8)=='file:///') f = f.substr(8);
				f = qs.unescape(f);
				var stats = fs.statSync(f);
				if (stats.isFile()){
					////new File(name, path, size)
					var name = f.split('/').pop();
					files.push( new File(name, f, stats.size) );
				}else if (stats.isDirectory()){
					fx.readDir(f, files);
				}
			});

			self.emit('uploadRequested', files);

		}else if (src=='ftpExplorer.tableWidget') {

			// FILES DRAGGED WITHIN FTPEXPLORER TABLE
			var tableItem = self.tableWidget.itemAt(e.pos());
			if (tableItem && tableItem.isFolder()){
				var targetPath = self.path + tableItem.text()+'/';
				self._maybeMoveSelectedItems(targetPath);
			}
		}

	});
}

/**
 * _setupComboBox
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._setupComboBox = function(){
	var self = this;

	self.tableComboBox.on('activated', function(index){
		self.changeDir(self.tableComboBox.itemText(index),function(){
			var items = self.path.split('/');
			items.shift();
			items.pop();
			self.currentTreeItem = self.rootItem;
			self.treeWidget.setCurrentItem(self.currentTreeItem);
			var newTreeItem;
			items.forEach(function(name){
				var cnt = self.currentTreeItem.childCount();
				for (var i=0;i<cnt;i++){
					if (self.currentTreeItem.child(i).text()==name){
						newTreeItem = self.currentTreeItem.child(i);
						break;
					}
				}
				self.currentTreeItem = newTreeItem;
				self.treeWidget.setCurrentItem(self.currentTreeItem);
			});
		});
	});
}

/**
 * _setupClient
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._setupClient = function(){
	var self = this;

	self.client.on('error', function(err) {
		if (err.code==421){//No transfer timeout (600 seconds): closing control connection

			self.connected = false;
			self.tableWidget.setAcceptDrops(false);
			self.tableWidget.setDragEnabled(false);
			self.emit('disconnected');

			self.tableWidget.setRowCount(0);
			self._resetTreeWidget();

			if (settings.autoReconnect){
				//console.log('Reconnecting...');//TODO
				self.connect(self.currentSite, false, function(){//self.path
					//console.log('succesfully reconnected');
				});
			}
		}else
			self.emit('error', err);
	});

//	self.client.on('greeting', function(msg) {console.log('FTP-MSG', msg)});
//	self.client.on('ready', function(msg) {console.log('FTP-READY')});
//	self.client.on('close', function(hadErr) {console.log('FTP-CLOSE', hadErr)});
//	self.client.on('end', function() {console.log('FTP-END')});
}

/**
 * _decode
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._decode = function(str){
	var self = this;

	switch(self.encoding){
	case 'utf-8':
		return (new Buffer(str, 'binary')).toString('utf-8');
		break;
	default:
		return str;
	}
}

/**
 * _getListRecursive
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._getListRecursive = function(path, files, folders, callback){
	var self = this;

	self.count++;
	self.client.list(path, function(err, list){
  	if (err) return self.emit('error', err);
		list.forEach(function(f){
			if (f.type=='d'){
				folders.push(path+f.name+'/');
				self._getListRecursive(path+f.name+'/',files, folders, callback);
			}else{
				files.push( new File(f.name, path+f.name, f.size, f.date) );
			}
		});
		self.count--;
		if (self.count == 0 && callback) callback();
  });
}

/**
 * _maybeDeleteSelectedItems
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._maybeDeleteSelectedItems = function(){
	var self = this;

	var items = self.tableWidget.selectedItems();
	if (items.length==0) return;
	if (settings.showConfirmDialog){
		var ok = items.length==3?
			qt.confirm('Delete selected item?') :
			qt.confirm('Delete selected items?'); // files/folders
		if (!ok) return;
	}
	items.forEach(function(item){
		if (item.column()==0){
			var name = item.text();
			if (name=='..') return;
			var f = self.path + name;
			if (item.isFolder()){
				self.deleteDirRecursive(f+'/');
			}else{
				self.deleteFile(f);
			}
		}
	});
}

/**
 * _maybeMoveSelectedItems
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._maybeMoveSelectedItems = function(targetPath){
	var self = this;

	var items = self.tableWidget.selectedItems();
	if (items.length==0) return;
	if (settings.showConfirmDialog){
		var ok = qt.confirm('Move selected item(s) to '+targetPath+'?');
		if (!ok) return;
	}
	items.forEach(function(item){
		if (item.column()==0){
			var name = item.text();
			if (name=='..') return;
			var f = self.path + name;
			self.renameFile(f, targetPath+name);
		}
	});
}

/**
 * Asks user for new name, if result is valid, renames the selected file/folder
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._maybeRenameSelectedItem = function(){
	var self = this;

//	var row = self.tableWidget.currentRow();
//	var item = self.tableWidget.item(row, 0);

	var items = self.tableWidget.selectedItems();
	if (items.length==0) return;
	var item = items[0];

	var isFolder = item.isFolder();
	var oldName = item.text();
	var name = qt.QInputDialog.getText('Rename '+(isFolder?'folder':'file'),
	    'Enter a new name:', oldName);
	if (typeof name!='string' || name=='' || name==oldName) return;
	if (settings.illegalChars){
		name = name.replace(new RegExp('['+settings.illegalChars+']', 'g'), '');
	}
	self.renameFile(self.path+oldName, self.path+name);
}

/**
 * Asks user for new folder name, if result is valid, creates new folder
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._maybeMakeDir = function(){
	var self = this;

	var name = qt.QInputDialog.getText('Create new folder', 'Enter the name:', '');
	if (typeof name!='string' || name=='') return;
	if (settings.illegalChars){
		name = name.replace(new RegExp('['+settings.illegalChars+']', 'g'), '');
	}
	this.makeDir(self.path+name);
}

/**
 * Resets (empties) the tree widget and sets the root text to the specied name
 *
 * @param {string} name - Text of treeWidget's root item.
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._resetTreeWidget = function(name){
	var self = this;

	self.treeWidget.setCurrentItem(self.rootItem);
	self.currentTreeItem = self.rootItem;
	var cnt = self.rootItem.childCount();
	for (var i=cnt-1;i>=0;i--){
		var node = self.rootItem.child(i);
		self.rootItem.removeChild(node);
		node.destroy();
	}

	// also reset combobox
	self.tableComboBox.clear();
}

/**
 * Returns path to current tree item (/foo/bar/)
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._getPathToCurrentTreeItem = function(){
	var self = this;

	var current = self.currentTreeItem;
	var items = [];
	while (current){
		items.splice(0,0,current.text());
		current = current.parent();
	}
	items.shift(); // remove rootItem ('/')
	var path = '/';
	items.forEach(function(item){
		path += item+'/';
	});

	return path; //items;
}

/**
 * Returns path to specified tree item (/foo/bar/)
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype._getPathToTreeItem = function(treeitem){
	var self = this;

	var current = treeitem;
	var items = [];
	while (current){
		items.splice(0,0,current.text());
		current = current.parent();
	}
	items.shift(); // remove rootItem ('/')
	var path = '/';
	items.forEach(function(item){
		path += item+'/';
	});
	return path; //items;
}

//################################################
// PUBLIC METHODS
//################################################

/**
 * connect (=async)
 *
 * @param {string|object} site
 * @param {string} startPath - Allows to overwrite site.rdir [optional]
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.connect = function(site, startPath, cb){
	var self = this;

	function _connect(){
		var cfg = typeof site=='string'?global.sites[site]:site;
		if (self._debug) cfg.debug = console.log;

		// set default
		self.encoding = 'binary'; // ???

	  self.client.once('ready', function() {
			self.connected = true;
			self.currentSite = site;

			self.tableWidget.setAcceptDrops(true);
			self.tableWidget.setDragEnabled(true);
			self.emit('connected');

			var initFunction = function(){
				if (!startPath) startPath = cfg.rdir;
				if (startPath==undefined||startPath==''||startPath=='/'){
					self.path = '/'; // ???
					self.update();
					if (cb) cb();
				}else{
					// create treeNodes for path (eg /foo/bar/)
					var dirs = startPath.split('/');
					if (dirs[0]=='') dirs.shift();
					if (dirs[dirs.length-1]=='') dirs.pop();
					// add according to folder list
					dirs.forEach(function(dir){
						var childTreeItem = new qt.QTreeWidgetItem;
						childTreeItem.setText(0, dir);
						self.currentTreeItem.addChild(childTreeItem);
						childTreeItem.setFlags(qt.ITEMFLAG.ITEMISENABLED|qt.ITEMFLAG.ITEMISSELECTABLE);
						childTreeItem.setIcon(0, __dirname, true);
						self.currentTreeItem.setExpanded(true);
						self.currentTreeItem = childTreeItem;
					});
		  		self.changeDir(startPath, cb);
		  	}
			}

			if (ut.isArray(self.client._feat) && self.client._feat.indexOf('UTF8')>=0){//supports UTF8
				self.client.raw('OPTS UTF8 ON', function(err, responseText, responseCode){
					if (err) {
						self.emit('error', err); // ignore
					}else{
						self.encoding = 'utf-8';
					}
					initFunction();
				});
			}else{
				initFunction();
			}
	  });
	  self.client.connect(cfg);
	}

	if (self.connected) self.disconnect(_connect);
	else _connect();
};

/**
 * disconnect (=async)
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.disconnect = function(cb){
	var self = this;
	self.client.once('end', function() {
		self.connected = false;
		self.currentSite = undefined;
		self.tableWidget.setAcceptDrops(false);
		self.tableWidget.setDragEnabled(false);
		self.emit('disconnected');

		self._resetTreeWidget();
		self.tableWidget.setRowCount(0);

		if (cb) cb();
	});
	self.client.end();
};

/**
 * getCurrentPath() -> string path
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.getCurrentPath = function(){
	//return this.getPath(this.currentTreeItem);
	return this.path;
}

/**
 * renameFile (path, newName) (=async)
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.renameFile = function(oldPath, newPath, cb){
	var self = this;

	self.client.rename(oldPath, newPath, function(err){
		if (err) return self.emit('error', err);
		self.update();
		if (cb) cb();
	});
}

/**
 * deleteFile (path) (=async)
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.deleteFile = function(path, cb){
	var self = this;

	self.client.delete(path, function(err){
		if (err) return self.emit('error', err);
		self.update();
		if (cb) cb();
	});
}

/**
 * changeDir (path) (=async)
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.changeDir = function(path, cb){
	var self = this;

	self.client.cwd(path, function(err){
		if (err) return self.emit('error', err);
		self.path = path;
		self.update();
		if (cb) cb();
	});
}

/**
 * makeDir (dir) (=async)
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.makeDir = function(path, cb){
	var self = this;

	self.client.mkdir(path, function(err, currentDir){
		if (err) return self.emit('error', err);
		self.update();
		if (cb) cb();
	});
}

/**
 * renameDir (dir, newName) (=async)
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.renameDir = function(oldPath, newPath, cb){
	var self = this;

	self.renameFile(oldPath, newPath, cb);
}

/**
 * deleteDir (dir) (=async)
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.deleteDir = function(path, cb){
	var self = this;

	self.client.rmdir(path, function(err, currentDir){
		if (err) return self.emit('error', err);
		self.update();
		if (cb) cb();
	});
}

/**
 * deleteDirRecursive (dir) (=async)
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.deleteDirRecursive = function(path, cb){ //path: ends on /
	var self = this;

	var files = [];
	var folders = [];
	self.count = 0;
	self._getListRecursive(path, files, folders, function(){
		var actions = [];
		files.forEach(function(f){
			actions.push(
				function(callback){
					self.client.delete(f.path, function(err){
						if (err) return self.emit('error', err);
						callback();
					});
				}
			);
		});
		folders.reverse();
		folders.push(path);
		folders.forEach(function(f){
			actions.push(
				function(callback){
					self.client.rmdir(f, function(err){
						if (err) return self.emit('error', err);
						callback();
					});
				}
			);
		});
		// START ACTION QUEUE
		async.series(actions, function(){
			self.update();
			if (cb) cb();
		});
	});
}

/**
 * update ()
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.update = function(){
	var self = this;

	self.client.list(function(err, list) {
	  if (err) return self.emit('error', err);

		// UPDATE TABLEWIDGET
	  list.push({name:'..',	type:'d',	size:0, date: undefined});

	  self.tableWidget.setRowCount(0); // TODO: release/destroy existing items
		self.tableWidget.setRowCount(list.length);

		var row = 0;
		var folders = [];
		var tableWidgetItem;

		self.tableWidget.setSortingEnabled(false);

		for (row=0;row<list.length;row++){
			var f = list[row];
			if (f.type=='d'){

				// FOLDER

				// NAME COLUMN
				tableWidgetItem = new qt.QTableWidgetItem;
				tableWidgetItem.setFolder(self._decode(f.name));
				tableWidgetItem.setIcon(__dirname, true);
				self.tableWidget.setItem(row, 0, tableWidgetItem);

				// SIZE COLUMN
				//tableWidgetItem = new qt.QTableWidgetItem;
				//tableWidgetItem.setText('');
				//self.tableWidget.setItem(row, 1, tableWidgetItem);

				// TYPE COLUMN
				tableWidgetItem = new qt.QTableWidgetItem;
				tableWidgetItem.setData(qt.ITEMDATAROLE.DISPLAYROLE, new qt.QVariant('Folder'));
				self.tableWidget.setItem(row, 2, tableWidgetItem);

				// DATE COLUMN
				if (f.date!=undefined){
					tableWidgetItem = new qt.QTableWidgetItem;
					tableWidgetItem.setData(qt.ITEMDATAROLE.DISPLAYROLE, new qt.QVariant(f.date));
					self.tableWidget.setItem(row, 3, tableWidgetItem);
				}

				if (f.name!='.' && f.name!='..') folders.push(f.name);

			}else{

				// FILE

				// NAME COLUMN
				tableWidgetItem = new qt.QTableWidgetItem;
				//tableWidgetItem.setFile(f.name);
				tableWidgetItem.setFile(self._decode(f.name));
				tableWidgetItem.setIcon(f.name, false);
				self.tableWidget.setItem(row, 0, tableWidgetItem);

				// SIZE COLUMN
				tableWidgetItem = new qt.QTableWidgetItem;
				tableWidgetItem.setData(qt.ITEMDATAROLE.DISPLAYROLE, new qt.QVariant(f.size));
				self.tableWidget.setItem(row, 1, tableWidgetItem);

				// TYPE COLUMN
				tableWidgetItem = new qt.QTableWidgetItem;
				tableWidgetItem.setData(qt.ITEMDATAROLE.DISPLAYROLE, new qt.QVariant(f.name.split('.').pop()+' File'));
				self.tableWidget.setItem(row, 2, tableWidgetItem);

				// DATE COLUMN
				tableWidgetItem = new qt.QTableWidgetItem;
				tableWidgetItem.setData(qt.ITEMDATAROLE.DISPLAYROLE, new qt.QVariant(f.date));
				self.tableWidget.setItem(row, 3, tableWidgetItem);
			}

		}

		self.tableWidget.setSortingEnabled(true);
		self.tableWidget.sortByColumn(0,0); // default: sort by name

		// UPDATE TREEWIDGET

		folders.sort();

		// remove existing childItems
		var cnt = self.currentTreeItem.childCount();
		for (var i=cnt-1;i>=0;i--){
			var node = self.currentTreeItem.child(i);
			self.currentTreeItem.removeChild(node);
			node.destroy();
		}

		// add according to folder list
		folders.forEach(function(folderName){
			var childTreeItem = new qt.QTreeWidgetItem;
			//childTreeItem.setData(0,3, childFolderNode.id)
			childTreeItem.setText(0, folderName);
			childTreeItem.setFlags(qt.ITEMFLAG.ITEMISENABLED|qt.ITEMFLAG.ITEMISSELECTABLE);
			childTreeItem.setIcon(0, __dirname, true);
			self.currentTreeItem.addChild(childTreeItem);
		});
		self.currentTreeItem.setExpanded(true);

	});
}

/**
 * Downloads currently selected (or passed as string list) files and folders
 * in current dir; emits 'downloadRequested'
 *
 * @param {string} itemNames - Array of file or folder names. [optional]
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.downloadSelection = function(itemNames, targetPath){
	var self = this;

	var items;
	if (itemNames==undefined){
		items = self.tableWidget.selectedItems();
	}else{
		items = [];
		itemNames.forEach(function(name){
			var found = self.tableWidget.findItems(name, 0);
			if (found.length) items.push(found[0]);
		});
	}

	if (items.length==0) return;

	var files = [];
	var folders = [];

	// selection can be list of files and folders (flat, i.e. only in current path)
	items.forEach(function(item){
		if (item.column()==0){
			var name = item.text();
			if (name=='..') return;
			var path = self.path + name;
			if (item.isFolder()){
				folders.push(path+'/');
			}else{
				//var size = self.tableWidget.item(item.row(), 1).getNumber();
				var size = self.tableWidget.item(item.row(), 1).data().toInt();
				files.push( new File(name, path, size) );
			}
		}
	});

	// for folders, we have to recursivley find all files inside
	var lookups = [];
	folders.forEach(function(f){
		lookups.push(function(cb){
			self.count = 0;
			self._getListRecursive(f, files, folders, cb);
		});
	});

	async.series(lookups, function(){
		self.emit('downloadRequested', files, self.path, targetPath);
	});

}

/**
 * Downloads currently selected (or passed as string list) files and folders
 * in current dir; emits 'downloadRequested'
 *
 * @param {string} itemNames - Array of file or folder names. [optional]
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.previewFile = function(name){
	var self = this;

	var src = self.path + name;
	var dest = process.env.TMP.replace(/\\/g, '/')+'/'+Math.floor(Math.random()*(999999-100000+1)+100000)+'-'+name;
	self.client.get(src, function(err, stream){
		if (err) {
			return self.emit('error', err);
		}
	  stream.once('end', function(){ //close
	  	cp.exec('start "" "'+dest+'"');
	  });
	  var ws = fx.createStreamAndFolders(dest); // creates local folders on the fly!
	  stream.pipe(ws); //fs.createWriteStream(dest)
	});
}

/**
 * Returns the container widget (a QSplitter in this case)
 *
 * @this {FTPExplorer}
 */
FTPExplorer.prototype.container = function(){
	var self = this;

	return self.hsplitter;
}

return FTPExplorer;
})();// END CLASS
