// ENCODING: UTF-8
'use strict';

// GLOBALS USED:
// - sites
// - settings

/**
 * fxFTP
 *
 * Comments in JSDoc 3 format (more or less)
 *
 * @file Core app file for fxFTP.
 * @author Valentin Schmidt
 * @copyright Valentin Schmidt 2013
 * @version 0.0.1
 *
 * @todo ...
 */
		
(function () {

process.chdir(__dirname);

// HERE PRIVATE STATIC VARS

// BUILT-IN MODULES
var cp = require('child_process');
var fs = require('fs');
var qs = require('querystring');

// EXTERNAL MODULES
var qt = require('./lib/node-qt');
require('./lib/dbg');

// APP MODULES
var SiteManager = require('./sitemanager');
var Explorer = require('./explorer');
var FtpExplorer = require('./ftpexplorer');
var TransferManager = require('./transfermanager');


/**
 * App Object
 *
 * @constructor
 */
var App = function(){
	var self = this;
	
	/**
	 * Called when user selected a site to open
	 *
	 * @callback
	 * @this {object} - QAction that was triggered by user!
	 */
	self.sitesSelected = function(){
		var name = this.text();
		self.remoteExplorer.connect(name);
	}

	/**
	 * Called when user tries to quit
	 *
	 * @callback
	 * @this {object} - QAction that was triggered by user!
	 */
	self.exit = function(e){
	  if (self.maybeExit()) {
	  	self.statesSave();
	  	self.settingsSave();
			process.exit();
	  } else if (e) e.ignore();
	}

	/**
	 * sitesChanged
	 *
	 * @callback
	 * @this {object} - SiteManager instance
	 */
	self.sitesChanged = function(){
		self.sitesSave();
		// update menu
		self.menus.siteslist.clear();
		for (var name in global.sites){
			self.menus.siteslist.addAction(name).on('triggered', self.sitesSelected);
		}	
	}

	//################################################
	// START
	//################################################
	
	self.appName = 'fxFTP';
	self.appOrg = 'fx';
	self.appVersion = '0.0.1'
	
	self.actions  = {};
	self.menus = {};
	
	// CREATE APP AND WIDGETS
	self.qtApp = new qt.QApplication;
	self.qtApp.setApplicationName(self.appName);
	self.qtApp.setOrganizationName(self.appOrg);
	self.qtApp.setApplicationVersion(self.appVersion);

	self.mainWindow = new qt.QMainWindow(0, true);
	self.menuBar = new qt.QMenuBar(self.mainWindow);
	self.toolBar = new qt.QToolBar(self.mainWindow);
	self.quickBar = new qt.QToolBar(self.mainWindow);
	self.statusBar = new qt.QStatusBar(self.mainWindow);
	self.vsplitter = new qt.QSplitter(self.mainWindow);
	self.hsplitter = new qt.QSplitter(self.mainWindow);
	
	self.messageLog = new qt.QPlainTextEdit(self.mainWindow);
	self.setupMessageLog();
	
	self.settingsLoad();
	self.sitesLoad();
	self.statesLoad();

	self.siteManager = new SiteManager(self.mainWindow);
	self.localExplorer = new Explorer(self.mainWindow);
	self.remoteExplorer = new FtpExplorer(self.mainWindow);
	self.transferManager = new TransferManager(self.mainWindow);

	// SETUP
	self.setupActions();
	self.setupMenus();
	self.setupToolBar();
	self.setupQuickBar();
	self.setupSplitters();
	self.setupSiteManager();
	self.setupLocalExplorer();
	self.setupRemoteExplorer();
	self.setupTransferManager();
	self.setupMainWindow();

	self.loadPlugins();
	self.mainWindow.show();
	
	self.localExplorer.loadFilesystem(); // might take a couple of seconds

	process.on('uncaughtException', function (err) {
		
		// show in messageLog
		self.messageLog.err('Uncaught exception: ' + err);
	
		// show in console
		console.err(err);
		
		// log to file;
		var str = err.stack + '\n';      
		fs.appendFileSync('./log/error.log', str, 'binary');	
		
	//	qt.alert('An unknown Error occured. Application needs to restart');
	//	var cmd = process.argv[0];
	//	process.execArgv.forEach(function (val) {cmd += ' '+val;});
	//	cmd += ' '+process.argv[1];
	//	cp.exec('start /B '+cmd);
	//	process.exit(1);
	
	});
	
	// JOIN NODE'S EVENT LOOP
	setInterval(self.qtApp.processEvents, 0);

};

/**
 * Loads modules found in 'plugins' folder.
 *
 * @this {App}
 */
App.prototype.loadPlugins = function(){
	var self = this;
	
	fs.readdir('./plugins', function(err, files){
		if (err) return self.messageLog.err(err);
		files.forEach(function(f){
			require('./plugins/'+f)(self);
		});
	})
}

/**
 * Checks for unsaved changes, allows to cancel Exit action.
 *
 * @this {App}
 */
App.prototype.maybeExit = function(){
	var self = this;
	
	if (self.transferManager.taskList.length>0){ // && transferManager.taskList.running
		var msgBox = new qt.QMessageBox(self.mainWindow);
		msgBox.setIcon(4);
		msgBox.setWindowTitle(self.appName);
		msgBox.setText('There are unfinished transfers in the transfer queue.\nExit anyway?');

		msgBox.addButton("Yes (Save queue)" , qt.BUTTONROLE.ACCEPTROLE); // 0
		msgBox.addButton("Yes (Discard)" , qt.BUTTONROLE.DESTRUCTIVEROLE); // 2
		msgBox.addButton("No" , qt.BUTTONROLE.REJECTROLE); // 1
		
		msgBox.setDefaultButton(0);
		var ret = msgBox.exec();
		
		if (ret == 0){
			self.transferManager.saveTasks();
			return true;
		}else if (ret == 1)
			return true;
		else //if (ret == 2)
			return false;
	}
	return true;
}

/**
 * settingsLoad
 *
 * @this {App}
 */
App.prototype.settingsLoad = function(){
	var self = this;
	
	try{
		global.settings = JSON.parse(fs.readFileSync('./config/settings.json', 'binary'));
	}catch(e){
		self.messageLog.err('Settings file missing or corrupted');
		global.settings = {};
		self.settingsSave();
	}
}

/**
 * settingsSave
 *
 * @this {App}
 */
App.prototype.settingsSave = function(){
	fs.writeFileSync('./config/settings.json', JSON.stringify(global.settings, null, '  '), 'binary');
}

/**
 * sitesLoad
 *
 * @this {App}
 */
App.prototype.sitesLoad = function(){
	var self = this;
	
	try{
		global.sites = JSON.parse(fs.readFileSync('./config/sites.json', 'binary'));
	}catch(e){
		self.messageLog.err('Sites file missing or corrupted');
		global.sites = {};
	}
}

/**
 * sitesSave
 *
 * @this {App}
 */
App.prototype.sitesSave = function(){
	fs.writeFileSync('./config/sites.json', JSON.stringify(global.sites, null, '  '), 'binary');
}

/**
 * statesLoad
 */
App.prototype.statesLoad = function(){
	try{
		global.state = JSON.parse(fs.readFileSync('./config/state.json', 'binary'));
	}catch(e){
		global.state = {};
	}
}

/**
 * statesSave
 *
 * @this {App}
 */
App.prototype.statesSave = function(){
	fs.writeFileSync('./config/state.json', JSON.stringify(global.state, null, '  '), 'binary');
}

/**
 * setupMessageLog
 *
 * @this {App}
 */
App.prototype.setupMessageLog = function(){
	var self = this;
	
	self.messageLog.setReadOnly(true);
	self.messageLog.log = function(str){this.appendPlainText(str);}
	self.messageLog.err = function(str){
		//this.appendPlainText('ERROR: '+str);
		this.appendHtml('<span style="color:red">'+str+'</span>');
	}
}

/**
 * setupActions
 *
 * @this {App}
 */
App.prototype.setupActions = function(){
	var self = this;
	
	// connect
	self.actions.connect = new qt.QAction('&Connect to...');
	self.actions.connect.setIcon(__dirname + '/assets/toolbar/connect.png');
	self.actions.connect.setShortcut('Ctrl+N');
	self.actions.connect.setStatusTip('Connect to Server...');

	// disconnect
	self.actions.disconnect = new qt.QAction('&Disconnect');
	self.actions.disconnect.on('triggered', function(){
		self.remoteExplorer.disconnect();
	});
	self.actions.disconnect.setIcon(__dirname + '/assets/toolbar/disconnect.png');
	self.actions.disconnect.setShortcut('Ctrl+W');
	self.actions.disconnect.setStatusTip('Close an existing connection');
	self.actions.disconnect.setEnabled(false);

	// exit
	self.actions.exit = new qt.QAction('E&xit');
	self.actions.exit.on('triggered', self.exit);
	//actions.exit.setIcon(__dirname + '/assets/toolbar/cancel.png')
	self.actions.exit.setShortcut('Ctrl+Q');
	self.actions.exit.setStatusTip('Exit the application');

	// OPTIONS
	
	// showGrid
	self.actions.showGrid = new qt.QAction('Show Grid');
	self.actions.showGrid.setCheckable(true);
	self.actions.showGrid.setChecked(global.settings.showGrid);
	self.actions.showGrid.on('toggled', function(checked){
		global.settings.showGrid = checked;		
		self.localExplorer.tableView.setShowGrid(global.settings.showGrid);
		self.remoteExplorer.tableWidget.setShowGrid(global.settings.showGrid);
		self.transferManager.queuedTransferTable.setShowGrid(global.settings.showGrid);
		self.transferManager.failedTransferTable.setShowGrid(global.settings.showGrid);
		self.transferManager.successfulTransferTable.setShowGrid(global.settings.showGrid);
	});
	// alternatingRowColors
	self.actions.alternatingRowColors = new qt.QAction('Alternating Row Colors');
	self.actions.alternatingRowColors.setCheckable(true);
	self.actions.alternatingRowColors.setChecked(global.settings.alternatingRowColors);
	self.actions.alternatingRowColors.on('toggled', function(checked){
		global.settings.alternatingRowColors = checked;		
		self.localExplorer.tableView.setAlternatingRowColors(global.settings.alternatingRowColors);
		self.remoteExplorer.tableWidget.setAlternatingRowColors(global.settings.alternatingRowColors);
		self.transferManager.queuedTransferTable.setAlternatingRowColors(global.settings.alternatingRowColors);
		self.transferManager.failedTransferTable.setAlternatingRowColors(global.settings.alternatingRowColors);
		self.transferManager.successfulTransferTable.setAlternatingRowColors(global.settings.alternatingRowColors);
	});
	//autoStartTransfers
	self.actions.autoStartTransfers = new qt.QAction('AutoStart Transfers');
	self.actions.autoStartTransfers.setCheckable(true);
	self.actions.autoStartTransfers.setChecked(global.settings.autoStartTransfers);
	self.actions.autoStartTransfers.on('toggled', function(checked){
		global.settings.autoStartTransfers = checked;		
	});
	//showConfirmDialog
	self.actions.showConfirmDialog = new qt.QAction('Show Confirm Dialogs');
	self.actions.showConfirmDialog.setCheckable(true);
	self.actions.showConfirmDialog.setChecked(global.settings.showConfirmDialog);
	self.actions.showConfirmDialog.on('toggled', function(checked){
		global.settings.showConfirmDialog = checked;		
	});
	//showHiddenFiles
	self.actions.showHiddenFiles = new qt.QAction('Show Hidden Files');
	self.actions.showHiddenFiles.setCheckable(true);
	self.actions.showHiddenFiles.setChecked(global.settings.showHiddenFiles);
	self.actions.showHiddenFiles.on('toggled', function(checked){
		global.settings.showHiddenFiles = checked;
		self.localExplorer.update();
	});
	//useRecycleBin
	self.actions.useRecycleBin = new qt.QAction('Use Recycle Bin');
	self.actions.useRecycleBin.setCheckable(true);
	self.actions.useRecycleBin.setChecked(global.settings.useRecycleBin);
	self.actions.useRecycleBin.on('toggled', function(checked){
		global.settings.useRecycleBin = checked;		
	});
	
	// STATE
	
	// showtoolbar
	self.actions.showtoolbar = new qt.QAction('ToolBar');
	self.actions.showtoolbar.setCheckable(true);
	self.actions.showtoolbar.setChecked(state.toolbar);
	self.actions.showtoolbar.on('toggled', function(checked){
		global.state.toolbar = checked;
		self.toolBar.setVisible(checked);
	});

	// showquickbar
	self.actions.showquickbar = new qt.QAction('QuickConnectBar');
	self.actions.showquickbar.setCheckable(true);
	self.actions.showquickbar.setChecked(state.quickbar);
	self.actions.showquickbar.on('toggled', function(checked){
		global.state.quickbar = checked;
		self.quickBar.setVisible(checked);
	});

	// showstatusbar
	self.actions.showstatusbar = new qt.QAction('StatusBar');
	self.actions.showstatusbar.setCheckable(true);
	self.actions.showstatusbar.setChecked(state.statusbar);
	self.actions.showstatusbar.on('toggled', function(checked){
		global.state.statusbar = checked;
		self.statusBar.setVisible(checked);
	});
	
	// showsitemanager
	self.actions.showsitemanager = new qt.QAction('SiteManager');
	self.actions.showsitemanager.setCheckable(true);
	self.actions.showsitemanager.setChecked(state.sitemanager);
	self.actions.showsitemanager.on('toggled', function(checked){
		//siteManager.container().setVisible(checked);
		global.state.sitemanager = checked;
		self.siteManager.container().setVisible(checked);
	});
	
	// showtransfermanager
	self.actions.showtransfermanager = new qt.QAction('TransferManager');
	self.actions.showtransfermanager.setCheckable(true);
	self.actions.showtransfermanager.setChecked(state.transfermanager);
	self.actions.showtransfermanager.on('toggled', function(checked){
		global.state.transfermanager = checked;
		self.transferManager.container().setVisible(checked);
	});
	
	// showlocaltree
	self.actions.showlocaltree = new qt.QAction('Local FolderTree');
	self.actions.showlocaltree.setCheckable(true);
	self.actions.showlocaltree.setChecked(state.localtree);
	self.actions.showlocaltree.on('toggled', function(checked){
		global.state.localtree = checked;
		self.localExplorer.treeView.setVisible(checked);
	});
	
	// showremotetree
	self.actions.showremotetree = new qt.QAction('Remote FolderTree');
	self.actions.showremotetree.setCheckable(true);
	self.actions.showremotetree.setChecked(state.remotetree);
	self.actions.showremotetree.on('toggled', function(checked){
		state.remotetree = checked;
		self.remoteExplorer.treeWidget.setVisible(checked);
	});
	
	// about
	self.actions.about = new qt.QAction('&About '+self.appName);
	self.actions.about.on('triggered', function(){
		qt.QMessageBox.about('About '+self.appName, '<b>'+self.appName+' v'+self.appVersion+'</b><br><br>'+
		'Free and Open Source FTP Client based on <a href="http://nodejs.org/">Node.js</a>, <a href="http://qt-project.org/doc/qt-4.8/">Qt 4.8</a> and a customized and extended version of the <a href="https://github.com/arturadib/node-qt#readme">Node-Qt Addon</a>.<br><br>This application is written in pure <a href="https://en.wikipedia.org/wiki/JavaScript">JavaScript</a> (<a href="http://en.wikipedia.org/wiki/ECMAScript">ECMA-262</a>, edition 5), and therefor can be customized or extended with any text editor, even Notepad.<br><br>License: <a href="http://en.wikipedia.org/wiki/WTFPL">WTFPL 2.0</a>');
	});

	// aboutnode
	self.actions.aboutnode = new qt.QAction('About Node.js'); // SHOW A CHM HELP FILE
	self.actions.aboutnode.on('triggered', function(){
		cp.exec('start "" "'+__dirname+'\\assets\\help\\Node.js-0.8.20.chm"');
	});

	// aboutqt
	self.actions.aboutqt = new qt.QAction('About Qt');
	self.actions.aboutqt.on('triggered', function(){
		qt.QMessageBox.aboutQt('About Qt');
	});

}

/**
 * setupMenus
 *
 * @this {App}
 */
App.prototype.setupMenus = function(){
	var self = this;
	
	self.menus.file = self.menuBar.addMenu('&File');
	self.menus.options = self.menuBar.addMenu('&Options');
	self.menus.view = self.menuBar.addMenu('&View');
	self.menus.help = self.menuBar.addMenu('&Help');

	// MENU FILE ###################

	self.menus.siteslist = self.menus.file.addMenu('Connect to...');
	for (var name in global.sites){
		self.menus.siteslist.addAction(name).on('triggered', self.sitesSelected);
	}
	self.menus.file.addAction(self.actions.disconnect);
	self.menus.file.addSeparator();
	self.menus.file.addAction(self.actions.exit);

	// MENU OPTIONS ###################
	
	self.menus.options.addAction(self.actions.showGrid);
	self.menus.options.addAction(self.actions.alternatingRowColors);
	self.menus.options.addAction(self.actions.autoStartTransfers);
	self.menus.options.addAction(self.actions.showConfirmDialog);
	self.menus.options.addAction(self.actions.showHiddenFiles);
	self.menus.options.addAction(self.actions.useRecycleBin);
	
	// MENU VIEW ###################

	self.menus.view.addAction(self.actions.showtoolbar);
	self.menus.view.addAction(self.actions.showquickbar);
	self.menus.view.addAction(self.actions.showstatusbar);
	self.menus.view.addSeparator();
	self.menus.view.addAction(self.actions.showsitemanager);
	self.menus.view.addAction(self.actions.showtransfermanager);
	self.menus.view.addSeparator();
	self.menus.view.addAction(self.actions.showlocaltree);
	self.menus.view.addAction(self.actions.showremotetree);

	// MENU HELP ###################

	self.menus.help.addAction(self.actions.aboutnode);
	self.menus.help.addAction(self.actions.aboutqt);
	self.menus.help.addSeparator();
	self.menus.help.addAction(self.actions.about);
}

/**
 * setupToolBar
 *
 * @this {App}
 */
App.prototype.setupToolBar = function(){
	var self = this;

	self.toolBar.setWindowTitle('ToolBar');
	self.toolBar.setIconSize(16,16);

	self.toolBar.on('visibilityChanged', function(vis){
		self.actions.showtoolbar.setChecked(vis);
	});

	// connect
	var toolButton = new qt.QToolButton(self.toolBar);
	toolButton.setDefaultAction(self.actions.connect);
	//toolButton.setIcon(__dirname + '/assets/toolbar/connect.png');
	toolButton.setMenu(self.menus.siteslist);
	toolButton.setPopupMode(2);//QToolButton::InstantPopup
	self.toolBar.addWidget(toolButton);

	// disconnect
	self.toolBar.addAction(self.actions.disconnect);
	self.toolBar.addSeparator();
	//self.toolBar.addAction(self.actions.exit);
}

/**
 * setupQuickBar
 *
 * @this {App}
 */
App.prototype.setupQuickBar = function(){
	var self = this;
	
	self.quickBar.setWindowTitle('Quick Connect Bar');

	var label, btn;

	label = new qt.QLabel(self.quickBar);
	label.setText(' Host: ');
	self.quickBar.addWidget(label);
	var serverLineEdit = new qt.QLineEdit;
	serverLineEdit.setFixedWidth(120);
	self.quickBar.addWidget(serverLineEdit);

	label = new qt.QLabel(self.quickBar);
	label.setText('   Port: ');
	self.quickBar.addWidget(label);
	var portLineEdit = new qt.QLineEdit;
	portLineEdit.setText('21');
	portLineEdit.setFixedWidth(30);
	self.quickBar.addWidget(portLineEdit);

	label = new qt.QLabel(self.quickBar);
	label.setText('   Username: ');
	self.quickBar.addWidget(label);
	var userLineEdit = new qt.QLineEdit;
	userLineEdit.setFixedWidth(80);
	self.quickBar.addWidget(userLineEdit);

	label = new qt.QLabel(self.quickBar);
	label.setText('   Password: ');
	self.quickBar.addWidget(label);
	var passLineEdit = new qt.QLineEdit;
	passLineEdit.setFixedWidth(80);
	passLineEdit.setEchoMode(qt.ECHOMODE.PASSWORD);
	self.quickBar.addWidget(passLineEdit);

	label = new qt.QLabel(self.quickBar);
	label.setText('   ');
	self.quickBar.addWidget(label);

	btn = new qt.QPushButton(self.quickBar);
	btn.setText('QuickConnect');
	btn.on('clicked', function(){
		var host = serverLineEdit.text();
		if (host.substr(0,6)=='ftp://') host = host.substr(6);
		if (host==''){
			self.messageLog.err('Host field empty!');
			return;
		}
		var config = {
		  host: host,
		  port: portLineEdit.text()*1,
		  user: userLineEdit.text(),
		  pass: passLineEdit.text()
		}
		self.remoteExplorer.connect(config);
	});
	btn.setDefault (true);
	self.quickBar.addWidget(btn);
	
	label = new qt.QLabel(self.quickBar);
	label.setText('   ');
	self.quickBar.addWidget(label);
	
	btn = new qt.QPushButton(self.quickBar);
	btn.setText('Save site...');
	
	btn.on('clicked', function(){
		var host = serverLineEdit.text();
		if (host.substr(0,6)=='ftp://') host = host.substr(6);
		var config = {
		  host: host,
		  port: portLineEdit.text()*1,
		  user: userLineEdit.text(),
		  pass: passLineEdit.text()
		}
		self.siteManager.showEditDialog('<New>', config);
	});

	self.quickBar.addWidget(btn);
}

/**
 * setupSplitters
 *
 * @this {App}
 */
App.prototype.setupSplitters = function(){
	var self = this;
	
	self.vsplitter.setOrientation(0);
	self.hsplitter.setOrientation(1);
	
	self.hsplitter.addWidget(self.localExplorer.container());
	self.hsplitter.addWidget(self.remoteExplorer.container());
	
	self.hsplitter.setCollapsible(0, false);
	self.hsplitter.setCollapsible(1, false);
	
	self.hsplitter.setSizes([350,350]);
	
	self.vsplitter.addWidget(self.messageLog);

	self.vsplitter.addWidget(self.hsplitter);
	self.vsplitter.addWidget(self.transferManager.container());

	self.vsplitter.setCollapsible(0, true);
	self.vsplitter.setCollapsible(1, false);
	self.vsplitter.setCollapsible(2, true);

	self.vsplitter.setSizes([100,350,150]);

	self.vsplitter.setStretchFactor(0,0);
	self.vsplitter.setStretchFactor(1,1);
	self.vsplitter.setStretchFactor(2,0);
}

/**
 * setupSiteManager
 *
 * @this {App}
 */
App.prototype.setupSiteManager = function(){
	var self = this;
	
	self.siteManager.on('connectRequested', function(site){
		self.remoteExplorer.connect(site);
	});
	
	self.siteManager.on('visibilityChanged', function(bool){
		self.actions.showsitemanager.setChecked(bool);
	});
	
	self.siteManager.on('sitesChanged', self.sitesChanged);
}

/**
 * setupLocalExplorer
 *
 * @this {App}
 */
App.prototype.setupLocalExplorer = function(){
	var self = this;
	
	// EVENT HANDLERS

	// forwarded to remoteExplorer to (asynchronously) find all files and subfolders (via FTP)
	// when done, result is passed to remoteExplorer.downloadRequested (emitted event)
	self.localExplorer.on('downloadRequested', function(itemNames, localPath){
		self.remoteExplorer.downloadSelection(itemNames, localPath);
	});

	self.localExplorer.on('uploadRequested', function(files){
		
		if (!self.remoteExplorer.connected){
			self.messageLog.err('Not connected to any host!');
			return;
		}

		var currentLocalPath = self.localExplorer.getCurrentPath();
		var currentRemotePath = self.remoteExplorer.getCurrentPath();

		var transfers = [];
		files.forEach(function(file){
			var transfer = {}; // transfer: {site, type, src, dest, size}
			transfer.site = self.remoteExplorer.currentSite;
			transfer.type = 'PUT';
			transfer.src = file.path;
			transfer.size = file.size;

			// dest = currentRemotePath + (path on client relative to currentLocalPath)
			var relativeRemotePath = file.path.substr(currentLocalPath.length);
			transfer.dest = currentRemotePath + relativeRemotePath;

			transfers.push(transfer);
		});

		self.transferManager.addTransfers(transfers);
	});

	self.localExplorer.on('error', function(err){
		self.messageLog.err(err);
	});
}

/**
 * setupRemoteExplorer
 *
 * @this {App}
 */
App.prototype.setupRemoteExplorer = function(){
	var self = this;

	//self.remoteExplorer._debug = true;
	
	self.remoteExplorer.on('connected', function(){
		self.actions.disconnect.setEnabled(true);
		var name = typeof self.remoteExplorer.currentSite=='string'?self.remoteExplorer.currentSite:'unnamed connection';
		self.mainWindow.setWindowTitle(self.appName+' - '+name);
	});

	self.remoteExplorer.on('disconnected', function(){
		self.actions.disconnect.setEnabled(false);
		self.mainWindow.setWindowTitle(self.appName);
	});

	self.remoteExplorer.on('error', function(err){
		self.messageLog.err(err);
	});

	// EVENT HANDLERS
	
	self.remoteExplorer.on('downloadRequested', function(files, currentRemotePath, localPath){
		if (!self.localExplorer.connected){
			self.messageLog.err('No local folder selected yet!');
			return;
		}

		if (!localPath) localPath = self.localExplorer.getCurrentPath();
		var transfers = [];
		files.forEach(function(file){
			var transfer = {}; // transfer: {site, type, src, dest, size}
			
			transfer.site = self.remoteExplorer.currentSite;
			transfer.type = 'GET';
			transfer.src = file.path;
			transfer.size = file.size;

			// currentLocalPath + (path on server relative to currentRemotePath)
			var relativeLocalPath = file.path.substr(currentRemotePath.length);
			transfer.dest = localPath + relativeLocalPath;

			transfers.push(transfer);
		});

		self.transferManager.addTransfers(transfers);
	});

	self.remoteExplorer.on('uploadRequested', function(files, remotePath){
		var currentLocalPath = self.localExplorer.getCurrentPath();
		if (!remotePath) remotePath = self.remoteExplorer.getCurrentPath();
		var transfers = [];
		files.forEach(function(file){
			var transfer = {}; // transfer: {site, type, src, dest, size}
			
			transfer.site = self.remoteExplorer.currentSite;
			transfer.type = 'PUT';
			transfer.src = file.path;
			transfer.size = file.size;

			var relativeRemotePath = file.path.substr(currentLocalPath.length);
			transfer.dest = remotePath + relativeRemotePath;
			transfers.push(transfer);
		});

		self.transferManager.addTransfers(transfers);
	});
}

/**
 * setupTransferManager
 *
 * @this {App}
 */
App.prototype.setupTransferManager = function(){
	var self = this;

	self.transferManager.on('done', function(){//just in case?
		self.remoteExplorer.update();
		self.localExplorer.update();
	});
	
	// forwarded to remoteExplorer to (asynchronously) find all files and subfolders (via FTP)
	// when done, result is passed to remoteExplorer.downloadRequested (emitted event)
	self.transferManager.on('downloadRequested', function(itemNames){
		var currentLocalPath = self.localExplorer.getCurrentPath();
		self.remoteExplorer.downloadSelection(itemNames, currentLocalPath);
	});
	
	
	self.transferManager.on('uploadRequested', function(files){
		
		if (!self.remoteExplorer.connected){
			self.messageLog.err('Not connected to any host!');
			return;
		}

		var currentLocalPath = self.localExplorer.getCurrentPath();
		var currentRemotePath = self.remoteExplorer.getCurrentPath();

		var transfers = [];
		files.forEach(function(file){
			var transfer = {}; // transfer: {site, type, src, dest, size}
			transfer.site = self.remoteExplorer.currentSite;
			transfer.type = 'PUT';
			transfer.src = file.path;
			transfer.size = file.size;

			// dest = currentRemotePath + (path on client relative to currentLocalPath)
			var relativeRemotePath = file.path.substr(currentLocalPath.length);
			transfer.dest = currentRemotePath + relativeRemotePath;

			transfers.push(transfer);
		});

		self.transferManager.addTransfers(transfers);
	});
	
	self.transferManager.on('error', function(err){
		self.messageLog.err(err);
	});
}

/**
 * setupMainWindow
 *
 * @this {App}
 */
App.prototype.setupMainWindow = function(){
	var self = this;
	
	self.mainWindow.setWindowIcon(__dirname + '/assets/icons/app.png');
	self.mainWindow.setMenuBar(self.menuBar);
	self.mainWindow.addToolBar(self.toolBar);
	self.mainWindow.addToolBar(self.quickBar);
	
	self.mainWindow.resize(1024, 768);
	self.mainWindow.setCentralWidget(self.vsplitter);
	self.mainWindow.setStatusBar(self.statusBar);
		
	self.mainWindow.addDockWidget(1, self.siteManager.container());
	
	// set visibility accourding to state
	self.toolBar.setVisible(global.state.toolbar);
	self.quickBar.setVisible(global.state.quickbar);
	self.statusBar.setVisible(global.state.statusbar);
	self.siteManager.container().setVisible(global.state.sitemanager);
	self.transferManager.container().setVisible(global.state.transfermanager);
	self.localExplorer.treeView.setVisible(global.state.localtree);
	self.remoteExplorer.treeWidget.setVisible(global.state.remotetree);
			
	self.mainWindow.setWindowTitle(self.appName);
	self.mainWindow.on('closeEvent', self.exit);

}

new App();

})();
