// ENCODING: UTF-8
'use strict';

// DEPENDS ON GLOBALS:
// - sites

// START SITEMANAGER CLASS
module.exports = (function () {

// HERE PRIVATE STATIC VARS

// BUILT-IN MODULES
var ev = require('events');
var fs = require('fs');
var ut = require('util');

// EXTERNAL MODULES
var fx = require('./lib/fx-utils');
var qt = require('./lib/node-qt');


/**
 * SiteManager Object
 * @constructor
 */
var SiteManager = function(parent){
	this._setup(parent);
}
ut.inherits(SiteManager, ev.EventEmitter);


//################################################
// PRIVATE METHODS
//################################################

/**
 * init
 *
 * @this {SiteManager}
 */
SiteManager.prototype._setup = function(parent){
	var self = this;

	// INSTANCE VARS
	self.dockWidget = new qt.QDockWidget(parent);
	self.listWidget = new qt.QListWidget(parent, true);
	self.editDialog = new qt.QDialog;
	self.containerWidget = new qt.QWidget(parent);
	self.buttons = {};
	
	self.oldName = false;

	self._setupDockWidget();
	self._setupListWidget();
	self._setupButtons();
	self._setupLayout();
	self._setupEditDialog();

}

/**
 * _setupDockWidget
 *
 * @this {SiteManager}
 */
SiteManager.prototype._setupDockWidget = function(){
	var self = this;

	self.dockWidget.setWindowTitle('SiteManager');
	self.dockWidget.setWidget(self.containerWidget);	
	self.dockWidget.setAllowedAreas(1 | 2);
	self.dockWidget.setFloating(false);
	self.dockWidget.setFixedWidth(150);

	// EVENT HANDLERS
	
	self.dockWidget.on('visibilityChanged', function(bool){
		self.emit('visibilityChanged', bool);
	});
}

/**
 * _setupListWidget
 *
 * @this {SiteManager}
 */
SiteManager.prototype._setupListWidget = function(){
	var self = this;

	self.listWidget.setSortingEnabled(true);
	for (var name in global.sites){
		self.listWidget.addItem(name);
	}
	
	// CREATE CONTEXT MENU
	
	self.listWidget.contextMenu = new qt.QMenu(self.listWidget);

	self.listWidget.contextMenu.addAction('Connect').on('triggered', function(){
		var listItem = self.listWidget.currentItem();
		var name = listItem.text();
		self.emit('connectRequested', name);
	});
	self.listWidget.contextMenu.addSeparator();
	self.listWidget.contextMenu.addAction('Edit Site').on('triggered', function(){		
		var item = self.listWidget.currentItem();
		var name = item.text();
		var site = global.sites[name];
		
		self.showEditDialog(name, site, true);
	});
	self.listWidget.contextMenu.addAction('Duplicate Site').on('triggered', function(){		
		var item = self.listWidget.currentItem();
		var name = item.text();
		var site = global.sites[name];
				
		self.showEditDialog(name+' Copy', site);
	});
	self.listWidget.contextMenu.addAction('Delete Site').on('triggered', function(){
		var ok = qt.confirm('Delete selected item?');
		if (ok){
			var item = self.listWidget.currentItem();
			var name = item.text();
			delete global.sites[name];
			item.destroy();
			//listWidget.setCurrentRow(0);
			self.emit('sitesChanged');
		}
	});
	
	// EVENT HANDLERS
	
	self.listWidget.on('itemDoubleClicked', function(listItem){
		var name = listItem.text();
		self.emit('connectRequested', name);
	});
	
	self.listWidget.on('contextMenuEvent', function(e){
		var listItem = self.listWidget.itemAt(e.pos());
		if (listItem){
			self.listWidget.contextMenu.exec(e.globalPos());
		}
	});
}

/**
 * _setupButtons
 *
 * @this {SiteManager}
 */
SiteManager.prototype._setupButtons = function(){
	var self = this;

	self.buttons.newBtn = new qt.QPushButton(self.containerWidget);
	self.buttons.newBtn.setText('New Site...');
	
	self.buttons.editBtn = new qt.QPushButton(self.containerWidget);
	self.buttons.editBtn.setText('Edit Site...');
	
	self.buttons.duplBtn = new qt.QPushButton(self.containerWidget);
	self.buttons.duplBtn.setText('Duplicate Site...');
	
	self.buttons.delBtn = new qt.QPushButton(self.containerWidget);
	self.buttons.delBtn.setText('Delete Site...');

	// EVENT HANDLERS
	
	self.buttons.newBtn.on('clicked', function(){
		self.showEditDialog('<New>', {});
	});
	
	self.buttons.editBtn.on('clicked', function(){
		if (self.listWidget.currentRow()<0) return;
		
		var item = self.listWidget.currentItem();
		var name = item.text();
		var site = global.sites[item.text()];
		self.showEditDialog(name, site, true);
	});
			
	self.buttons.duplBtn.on('clicked', function(){
		if (self.listWidget.currentRow()<0) return;
		
		var item = self.listWidget.currentItem();
		var name = item.text();
		var site = global.sites[name];
		self.showEditDialog(name+' Copy', site);
	});

	self.buttons.delBtn.on('clicked', function(){
		if (self.listWidget.currentRow()<0) return;
		
		var ok = qt.confirm('Delete selected item?');
		if (ok){
			var item = self.listWidget.currentItem();
			delete global.sites[item.text()];
			item.destroy();
			self.emit('sitesChanged');
		}
	});
}

/**
 * _setupLayout
 *
 * @this {SiteManager}
 */
SiteManager.prototype._setupLayout = function(){
	var self = this;

	var vlayout = new qt.QVBoxLayout();
	vlayout.setSpacing(2);
	vlayout.setContentsMargins(0,0,0,0);
	vlayout.addWidget(self.listWidget);
	vlayout.addWidget(self.buttons.newBtn);
	vlayout.addWidget(self.buttons.editBtn);
	vlayout.addWidget(self.buttons.duplBtn);
	vlayout.addWidget(self.buttons.delBtn);
	
	self.containerWidget.setLayout(vlayout);
}

/**
 * _setupEditDialog
 *
 * @this {SiteManager}
 */
SiteManager.prototype._setupEditDialog = function(){
	var self = this;

	var editDialog = self.editDialog;
	
	editDialog.setWindowTitle('Edit Site');
	editDialog.setWindowFlags(qt.WINDOWTYPE.CUSTOMIZEWINDOWHINT | qt.WINDOWTYPE.WINDOWTITLEHINT);// no close btn
	
	var label;
	
	var vlayout = new qt.QVBoxLayout();
	vlayout.setSizeConstraint( 3 ); // QLayout::SetFixedSize
	//vlayout.setSpacing(2);
	//vlayout.setContentsMargins(0,0,0,0);

	label = new qt.QLabel(self.editDialog);
	label.setText('Site Name: *');
	vlayout.addWidget(label);
	editDialog.editName = new qt.QLineEdit(self.editDialog);
	//edit1.setFixedWidth(250);
	//editName.on('textEdited', function(){});
	vlayout.addWidget(editDialog.editName);
	
	label = new qt.QLabel(self.editDialog);
	label.setText('Host: *');
	vlayout.addWidget(label);
	editDialog.editHost = new qt.QLineEdit(self.editDialog);
	vlayout.addWidget(editDialog.editHost);
	
	label = new qt.QLabel(self.editDialog);
	label.setText('Port: *');
	vlayout.addWidget(label);
	editDialog.editPort = new qt.QLineEdit(self.editDialog);
	vlayout.addWidget(editDialog.editPort);
	
	label = new qt.QLabel(self.editDialog);
	label.setText('Username:');
	vlayout.addWidget(label);
	editDialog.editUser = new qt.QLineEdit(self.editDialog);
	vlayout.addWidget(editDialog.editUser);
	
	label = new qt.QLabel(self.editDialog);
	label.setText('Password:');
	vlayout.addWidget(label);
	editDialog.editPass = new qt.QLineEdit(self.editDialog);
	editDialog.editPass.setEchoMode(qt.ECHOMODE.PASSWORD);
	vlayout.addWidget(editDialog.editPass);
	
	label = new qt.QLabel(self.editDialog);
	label.setText('Path:');
	vlayout.addWidget(label);
	editDialog.editPath = new qt.QLineEdit(self.editDialog);
	vlayout.addWidget(editDialog.editPath);
	
	vlayout.addSpacing(10);
	
	var btnBox = new qt.QDialogButtonBox(self.editDialog);
	var saveBtn = btnBox.addButton(qt.STANDARDBUTTON.SAVE);
	var discardBtn = btnBox.addButton(qt.STANDARDBUTTON.DISCARD);
	saveBtn.setDefault(true);
	vlayout.addWidget(btnBox);
	
	editDialog.setLayout(vlayout);
	
	// EVENT HANDLERS
	
	saveBtn.on('clicked', function(){
		var name = editDialog.editName.text();
		if (name==''||editDialog.editHost.text()==''||editDialog.editPort.text()==''){
			qt.alert('Please fill in all required fields!', 'Notice');
			return;
		}
		if (global.sites[name]!=undefined && name!=self.oldName){
			qt.alert('A site with this name already exists! Please choose a different name!', 'Notice');
			return;
		}
		
		var site = {
			host: editDialog.editHost.text(),
			port: editDialog.editPort.text() || 21
		}
		if (editDialog.editUser.text()!='') site.user = editDialog.editUser.text();
		if (editDialog.editPass.text()!='') site.pass = editDialog.editPass.text();
		if (editDialog.editPath.text()!='') site.rdir = editDialog.editPath.text();

		editDialog.close();
		
		if (self.oldName){ // site was edited
			if (name==self.oldName){ // not renamed
				global.sites[name] = site;
			}else{

				delete global.sites[self.oldName];
				global.sites[name] = site;
				global.sites = fx.sortByKey(global.sites); // re-sort
				
				// findItems ?
				var item = self.listWidget.currentItem();
				item.setText(name);
				
			}
		}else{ // new site
			global.sites[name] = site;
			global.sites = fx.sortByKey(global.sites); // re-sort
			
			self.listWidget.addItem(name);
		}
				
		self.emit('sitesChanged');
	});
	
	discardBtn.on('clicked', function(){
		// confirm?
		editDialog.close();
	});

}


//################################################
// PUBLIC METHODS
//################################################

/**
 * showEditDialog
 * Adds new site and saves site list
 *
 * @this {SiteManager}
 */
SiteManager.prototype.showEditDialog = function(name, site, editFlag){
	var self = this;
		
	self.oldName = editFlag ? name : false;
	self.editDialog.editName.setText(name);
	self.editDialog.editHost.setText(site.host || '');
	self.editDialog.editPort.setText(site.port || '21');
	self.editDialog.editUser.setText(site.user || '');
	self.editDialog.editPass.setText(site.pass || '');
	self.editDialog.editPath.setText(site.rdir || '');
	self.editDialog.exec();
}

/**
 * updateList
 *
 * @this {SiteManager}
 */
SiteManager.prototype.updateList = function(){
	var self = this;
	
	self.listWidget.clear();
	for (var name in global.sites){
		self.listWidget.addItem(name);
	}
}
	
/**
 * Returns the container widget
 *
 * @this {SiteManager}
 */
SiteManager.prototype.container = function(){
	var self = this;
	
	return self.dockWidget;
}

return SiteManager;
})();// END CLASS
