// ENCODING: UTF-8
'use strict';

// DEPENDS ON GLOBALS:
// - settings
// - sites

// START TRANSFERMANAGER CLASS
module.exports = (function () {

// HERE PRIVATE STATIC VARS

// BUILT-IN MODULES
var ev = require('events');
var fs = require('fs');
var qs = require('querystring');
var ut = require('util');

// EXTERNAL MODULES
var fx = require('./lib/fx-utils');
var qt = require('./lib/node-qt');
var FtpClient = require('./lib/ftp');

var File = require('./lib/file');

/**
 * TransferManager Object
 *
 * @constructor
 */
var TransferManager = function(parent){
	this._setup(parent);
}
ut.inherits(TransferManager, ev.EventEmitter);


//################################################
// PRIVATE METHODS
//################################################

/**
 * init
 *
 * @this {TransferManager}
 */
TransferManager.prototype._setup = function(parent){
	var self = this;

	// INSTANCE VARS
	self.tabWidget = new qt.QTabWidget(parent); // container
	self.queuedTransferTable = new qt.QTableWidget(parent, true);
	self.failedTransferTable = new qt.QTableWidget(parent, true);
	self.successfulTransferTable = new qt.QTableWidget(parent, true);

	self.taskList = [];
	self.running = false;

	self.connected = false;
	self.currentSite = undefined; // string (prop name of global.sites)

	//test
	self.currentWriteStream = false;

	self.client = new FtpClient();
	self.tmpClient = new FtpClient(); // only used to show upload progress
	self.uploadTimer = undefined;

	self._setupTabWidget();
	self._setupTables();

	self._setupClient();

	self.loadTasks();
}

/**
 * _setupTabWidget
 *
 * @this {TransferManager}
 */
TransferManager.prototype._setupTabWidget = function(){
	var self = this;

	self.tabWidget.setStyleSheet('QTabBar::tab:!selected {color:#666666;}');
	//this.tabWidget.setStyleSheet('QTabBar::tab:selected {color:#000099;}');
	//QTabBar::setTabTextColor ( int index, const QColor & color )

	self.tabWidget.setTabPosition(qt.TABPOSITION.SOUTH);
	self.tabWidget.addTab(self.queuedTransferTable, '&Queued transfers');
	self.tabWidget.addTab(self.failedTransferTable, '&Failed transfers');
	self.tabWidget.addTab(self.successfulTransferTable, '&Successful transfers');
}

/**
 * _setupTables
 *
 * @this {TransferManager}
 */
TransferManager.prototype._setupTables = function(){
	var self = this;

	self.queuedTransferTable.setObjectName('siteManager.queuedTransferTable');
	self.queuedTransferTable.setShowGrid(settings.showGrid);
	self.queuedTransferTable.setAlternatingRowColors(settings.alternatingRowColors);
	self.queuedTransferTable.setRowDefaultHeight(17);
	self.queuedTransferTable.setColumnCount(5);
	self.queuedTransferTable.setVerticalHeaderVisible(false);
	self.queuedTransferTable.setHorizontalHeaderLabels(['Site','Type','Source','Destination',  'Status']);
	self.queuedTransferTable.setHorizontalStretchLastSection(true);
	self.queuedTransferTable.setHorizontalHeaderDefaultAlignment(1);
	self.queuedTransferTable.setEditTriggers(0);
	self.queuedTransferTable.setSelectionBehavior(1);
	self.queuedTransferTable.setColumnWidth(0,50); //new: site
	self.queuedTransferTable.setColumnWidth(1,50);
	self.queuedTransferTable.setColumnWidth(2,300);
	self.queuedTransferTable.setColumnWidth(3,300);
	self.queuedTransferTable.setColumnWidth(4,50);

	self.queuedTransferTable.setDragDropMode(qt.DRAGDROPMODE.DRAGDROP);//DRAGDROP INTERNALMOVE
	self.queuedTransferTable.setDragEnabled(true);
	//self.queuedTransferTable.setVerticalHeaderMovable(true);

	[self.failedTransferTable, self.successfulTransferTable].forEach(function(table){
		table.setShowGrid(settings.showGrid);
		table.setAlternatingRowColors(settings.alternatingRowColors);
		table.setRowDefaultHeight(17);
		table.setColumnCount(4);
		table.setVerticalHeaderVisible(false);
		table.setHorizontalHeaderLabels(['Site','Type','Source','Destination']);
		table.setHorizontalStretchLastSection(true);
		table.setHorizontalHeaderDefaultAlignment(1);
		table.setEditTriggers(0);
		table.setSelectionBehavior(1);
		table.setColumnWidth(0,50);
		table.setColumnWidth(1,50);
		table.setColumnWidth(2,300);
		table.setColumnWidth(3,350);
	});

	// CREATE CONTEXT MENUS

	//################################################
	// queuedTransferTable
	//################################################
	self.queuedTransferTable.contextMenu = new qt.QMenu(self.queuedTransferTable);

	self.queuedTransferTable.contextMenu.addAction('Start transfer queue').on('triggered', function(){
		self.start();
	});
	self.queuedTransferTable.contextMenu.addAction('Stop transfer queue').on('triggered', function(){
		self.stop();
	});
	self.queuedTransferTable.contextMenu.addSeparator();

	self.queuedTransferTable.removeAction = self.queuedTransferTable.contextMenu.addAction('Remove selected tasks');
	self.queuedTransferTable.removeAction.on('triggered', function(){
		self.removeSelectedTransfers();
	});

	self.queuedTransferTable.clearAction = self.queuedTransferTable.contextMenu.addAction('Clear list');
	self.queuedTransferTable.clearAction.on('triggered', function(){
		self.clear();
	});

	self.queuedTransferTable.on('contextMenuEvent', function(e){
		if (self.queuedTransferTable.rowCount()==0) return;
		var items = self.queuedTransferTable.selectedItems();
		self.queuedTransferTable.removeAction.setEnabled(items.length>0);
		self.queuedTransferTable.contextMenu.exec(e.globalPos());
	});

	self.queuedTransferTable.on('dropEvent', function(e){

		switch (e.source()){

		case 'siteManager.queuedTransferTable': // MOVE ROW IN TABLE
			var srcRow = self.queuedTransferTable.currentRow();
			var destItem = self.queuedTransferTable.itemAt(e.pos());
			var destRow = destItem?destItem.row():self.queuedTransferTable.rowCount()-1;
			if (srcRow==destRow) return;
			if (self.running && srcRow==0){
				return self.emit('error', 'Running transfers can\'t be moved anymore');
			}
			var t = self.taskList.splice(srcRow, 1)[0];
			self.queuedTransferTable.removeRow(srcRow);
			self.queuedTransferTable.insertRow(destRow);
			var tableWidgetItem;
			tableWidgetItem = new qt.QTableWidgetItem;
			tableWidgetItem.setText(typeof t.site=='string'?t.site:'');
			self.queuedTransferTable.setItem(destRow, 0, tableWidgetItem);

			tableWidgetItem = new qt.QTableWidgetItem;
			tableWidgetItem.setText(t.type); // type: GET or PUT
			self.queuedTransferTable.setItem(destRow, 1, tableWidgetItem);

			tableWidgetItem = new qt.QTableWidgetItem;
			tableWidgetItem.setText(t.src);
			self.queuedTransferTable.setItem(destRow, 2, tableWidgetItem);

			tableWidgetItem = new qt.QTableWidgetItem;
			tableWidgetItem.setText(t.dest);
			self.queuedTransferTable.setItem(destRow, 3, tableWidgetItem);

			tableWidgetItem = new qt.QTableWidgetItem;
			tableWidgetItem.setText('0%');
			self.queuedTransferTable.setItem(destRow, 4, tableWidgetItem);
			self.taskList.splice(destRow,0,t);
			break;

		case 'explorer.tableView': // FILES/FOLDER DROPPED FROM LOCAL EXPLORER TABLE
			var data = e.mimeData(true)['text/uri-list'];
			if (data==undefined) return;
			var items = data.split('\r\n');
			items.pop();
			var files = [];
			items.forEach(function(f){
				if (f.substr(0,8)=='file:///') f = f.substr(8);
				f = qs.unescape(f);
				var stats = fs.statSync(f);
				if (stats.isFile()){
					var name = f.split('/').pop();
					files.push( new File(name, f, stats.size) );
				}else if (stats.isDirectory()){
					fx.readDir(f, files);
				}
			});

			self.emit('uploadRequested', files);
			break;

		case 'ftpExplorer.tableWidget': // FILES/FOLDER DROPPED FROM REMOTE EXPLORER TABLE
			var data = e.mimeData(true)['text/table-cells']; //text/plain
			if (data==undefined) return;
			var cells = data.split('\r\n');
			cells.pop();
			var lineCnt = cells.length/3;
			var itemNames = [];
			for (var i=0;i<lineCnt;i++){
				itemNames.push(cells[i]);
			}
			// HERE CONFIRM DIALOG?
			self.emit('downloadRequested', itemNames); // BOTH FILES AND FOLDERS!
			break;

		// TODO: ALSO HANDLE DROPS FROM EXPLORER?

		}//end switch

	});


	//################################################
	// successfulTransferTable
	//################################################

	self.successfulTransferTable.contextMenu = new qt.QMenu(self.successfulTransferTable);
	self.successfulTransferTable.contextMenu.addAction('Clear list').on('triggered', function(){
		self.successfulTransferTable.setRowCount(0);
	});
	self.successfulTransferTable.on('contextMenuEvent', function(e){
		if (self.successfulTransferTable.rowCount()==0) return;
		self.successfulTransferTable.contextMenu.exec(e.globalPos());
	});

	//################################################
	// failedTransferTable
	//################################################
	self.failedTransferTable.contextMenu = new qt.QMenu(self.failedTransferTable);

	self.failedTransferTable.retryAction = self.failedTransferTable.contextMenu.addAction('Retry selected tasks')
	self.failedTransferTable.retryAction.on('triggered', function(){
		var items = self.failedTransferTable.selectedItems();
		var tasks = [];
		items.forEach(function(item){
			if (item.column()==0){
				var row = item.row();
				self.queuedTransferTable.removeRow(row);
				var task = {};
				task.site = self.failedTransferTable.item(row, 0).text();
				task.type = self.failedTransferTable.item(row, 1).text();
				task.src  = self.failedTransferTable.item(row, 2).text();
				task.dest = self.failedTransferTable.item(row, 3).text();
				tasks.push(task);
			}
		});
		if (tasks.length) self.addTransfers(tasks);
	});
	self.failedTransferTable.contextMenu.addSeparator();
	self.failedTransferTable.contextMenu.addAction('Clear list').on('triggered', function(){
		self.failedTransferTable.setRowCount(0);
	});

	self.failedTransferTable.on('contextMenuEvent', function(e){
		if (self.failedTransferTable.rowCount()==0) return;
		var items = self.failedTransferTable.selectedItems();
		self.failedTransferTable.retryAction.setEnabled(items.length>0);

		self.failedTransferTable.contextMenu.exec(e.globalPos());
	});
}

/**
 * _setupClient
 *
 * @this {TransferManager}
 */
TransferManager.prototype._setupClient = function(){
	var self = this;

	self.client.on('error', function(err) {self.emit('error', err);});
	
//	self.client.on('greeting', function(msg) {log('FTP-MSG', msg)});
//	self.client.on('ready', function(msg) {log('FTP-READY')});
//	self.client.on('close', function(hadErr) {log('FTP-CLOSE', hadErr)});
//	self.client.on('end', function() {log('FTP-END')});
}

/**
 * _connect (=async)
 *
 * @param {string} siteName
 * @param {function} cb - Callback [optional]
 * @this {TransferManager}
 */
TransferManager.prototype._connect = function(siteName, cb){
	var self = this;

	function _connect(){
	  self.client.once('ready', function() {
			self.connected = true;
			self.currentSite = siteName;
			if (cb) cb();
	  });
	  self.client.connect(global.sites[siteName]);
	}

	if (self.connected) self._disconnect(_connect);
	else _connect();
};

/**
 * _disconnect (=async)
 *
 * @param {function} cb - Callback [optional]
 * @this {TransferManager}
 */
TransferManager.prototype._disconnect = function(cb){
	var self = this;
	self.client.once('end', function() {
		self.connected = false;
		self.currentSite = undefined;
		if (cb) cb();
	});
	self.client.end();
};

/**
 * _nextTask
 *
 * @this {TransferManager}
 */
TransferManager.prototype._nextTask = function(){
	var self = this;

	if (self.taskList.length==0){
		if (self.connected)	self._disconnect();
		self.running = false; // done.
		self.emit('done');
	}
	if (!self.running) return;

	var task = self.taskList[0];
	var statusItem = self.queuedTransferTable.item(0, 4); // item(row, col)

	// .executeTask(task, doneCb, progressCb);
	// * @param {object} task - props: site, type, src, dest, size. [optional]

	var doneCb = function(ok){

		self.taskList.shift();
		self.queuedTransferTable.removeRow(0);

		// add to successfulTransferTable or failedTransferTable
		var table = ok ? self.successfulTransferTable : self.failedTransferTable;
		var row = table.rowCount();
		table.setRowCount(row+1);

		var tableWidgetItem0 = new qt.QTableWidgetItem;
		tableWidgetItem0.setText(typeof task.site=='string'?task.site:'');
		table.setItem(row, 0, tableWidgetItem0);

		var tableWidgetItem0 = new qt.QTableWidgetItem;
		tableWidgetItem0.setText(task.type); // GET or PUT
		table.setItem(row, 1, tableWidgetItem0);

		var tableWidgetItem2 = new qt.QTableWidgetItem;
		tableWidgetItem2.setText(task.src);
		table.setItem(row, 2, tableWidgetItem2);

		var tableWidgetItem3 = new qt.QTableWidgetItem;
		tableWidgetItem3.setText(task.dest);
		table.setItem(row, 3, tableWidgetItem3);

		//if (self.running) callback(null);
		self._nextTask();
	};

	var progressCb = function(prog){
		statusItem.setText(Math.round(prog*100)+'%');
	};

	var taskType = task.type=='GET'?'_downloadFile':'_uploadFile';
	var func = function(){
		self[taskType](task.src, task.dest, task.size, doneCb, progressCb);
	};
	if (task.site==self.currentSite){
		func();
	}else{

		//self.connect(task.site, undefined, f);
		self.client.once('ready', function() {
			self.connected = true;
			self.currentSite = task.site;
			func();
		});
		self.client.connect(global.sites[task.site]);
	}
}

/**
 * _downloadFile (src, dest, progressCb) (=async)
 *
 * @this {TransferManager}
 */
TransferManager.prototype._downloadFile = function(src, dest, size, doneCb, progressCb){
	var self = this;

	if (size===false){
		// TODO: get size
	}
	self.client.get(src, function(err, stream){
		if (err) {
			if (doneCb) doneCb.call(null, false);
			return self.emit('error', err);
		}
		stream.size = 0;
	  stream.once('end', function(){ //close
	  	if (progressCb) progressCb.call(null, 1);
	  	if (doneCb) doneCb.call(null, true);
	  });
	  stream.on('data', function(chunk){
	  	stream.size += chunk.length;
	  	if (progressCb) progressCb.call(null, stream.size/size);
	  });
	  self.currentWriteStream = fx.createStreamAndFolders(dest); // creates local folders on the fly!
	  stream.pipe(self.currentWriteStream); //fs.createWriteStream(dest)
	});
}

/**
 * _uploadFile (src, dest, progressCb) (=async)
 *
 * @this {TransferManager}
 */
TransferManager.prototype._uploadFile = function(src, dest, size, doneCb, progressCb){
	var self = this;

	var startUpload = function(){

		if (progressCb){
			self.tmpClient.connect(global.sites[self.currentSite]);
		  self.tmpClient.on('ready', function() {
		    self.uploadTimer = setInterval(function(){
		    	self.tmpClient.size(dest, function(err, numBytes){
		    		if (err) self.emit('error', err);
		    		progressCb.call(null, numBytes/size);
		    	});
		    }, 2000);
		  });
	  }

		self.client.put(src, dest, function(err){

			if (progressCb) {
				if (!err) progressCb.call(null, 1);
				if (self.tmpClient.connected) self.tmpClient.end();
				clearInterval(self.uploadTimer);
				self.uploadTimer = undefined;
			}

			if (err){
				if (doneCb) doneCb.call(null, false);
				return self.emit('error', err);
			}
			if (doneCb) doneCb.call(null, true);
		});

	}

	var pos = dest.lastIndexOf('/');
	if (pos>0){
		// first create required folders
		var path = dest.substr(0, pos);
		// mkdir(< string >path, [< boolean >recursive, ]< function >callback)
		self.client.mkdir(path, true, function(err){
			startUpload();
		});
	}else{
		startUpload();
	}
}


//################################################
// PUBLIC METHODS
//################################################

/**
 * Appends specified transfers to the end of the transfer list
 *
 * @this {TransferManager}
 * @param {array} transfers (transfer = {site, type, src, dest, size})
 */
TransferManager.prototype.addTransfers = function(transfers){
	var self = this;

	var row = this.queuedTransferTable.rowCount();
	//this.queuedTransferTable.setRowCount(0); // TODO: release/destroy existing items
	this.queuedTransferTable.setRowCount(row+transfers.length);
	var tableWidgetItem;

	transfers.forEach(function(t){

		tableWidgetItem = new qt.QTableWidgetItem;
		tableWidgetItem.setText(typeof t.site=='string'?t.site:'');
		self.queuedTransferTable.setItem(row, 0, tableWidgetItem);

		tableWidgetItem = new qt.QTableWidgetItem;
		tableWidgetItem.setText(t.type); // type: GET or PUT
		self.queuedTransferTable.setItem(row, 1, tableWidgetItem);

		tableWidgetItem = new qt.QTableWidgetItem;
		tableWidgetItem.setText(t.src);
		self.queuedTransferTable.setItem(row, 2, tableWidgetItem);

		tableWidgetItem = new qt.QTableWidgetItem;
		tableWidgetItem.setText(t.dest);
		self.queuedTransferTable.setItem(row, 3, tableWidgetItem);

		tableWidgetItem = new qt.QTableWidgetItem;
		tableWidgetItem.setText('0%');
		self.queuedTransferTable.setItem(row, 4, tableWidgetItem);

		self.taskList[row] = t;

		row++;
	});

	if (settings.autoStartTransfers && !this.running) this.start();
}

/**
 * Stops and clears the transfer list
 *
 * @this {TransferManager}
 */
TransferManager.prototype.clear = function(){
	var self = this;

	if (self.running){

		if (self.uploadTimer){
			clearInterval(self.uploadTimer);
			self.uploadTimer = undefined;
		}

		if (self.tmpClient.connected){
			self.tmpClient.abort(true, function(){
				//console.log('tmpClient aborted');
			});
		}

		self.client.abort(true, function(){
			if (self.taskList.length && self.taskList[0].type=='GET'){
				if (self.currentWriteStream && self.currentWriteStream.writable){
					self.currentWriteStream.destroy();
					self.currentWriteStream = false;
				}
				var fn = self.taskList[0].dest;
				if (fs.existsSync(fn)){
					fs.unlinkSync(fn);
				}
			}

			self.taskList = [];
			self.queuedTransferTable.setRowCount(0);

			self.running = false;
			self.connected = false;
			self.currentSite = undefined;
		});

	}else{
		self.taskList = [];
		self.queuedTransferTable.setRowCount(0);
	}

}

/**
 * Removes selected transfers
 *
 * @this {TransferManager}
 */
TransferManager.prototype.removeSelectedTransfers = function(){
	var self = this;

	var items = self.queuedTransferTable.selectedItems();
	var rows = [];
	items.forEach(function(item){
		if (item.column()==0) rows.push(item.row());
	});
	if (rows.length==0) return;
	rows.sort(function(a,b){return a-b;});
	if (self.running && rows[0]==0){
		self.client.abort(true, function(){
			self.connected = false;
			self.currentSite = undefined; // string (prop name of global.sites)
			if (self.taskList[0].type=='GET'){
				if (self.currentWriteStream && self.currentWriteStream.writable){
					self.currentWriteStream.destroy();
					self.currentWriteStream = false;
				}
				var fn = self.taskList[0].dest;
				if (fs.existsSync(fn)){
					fs.unlinkSync(fn);
				}
			}
			for (var i=rows.length-1;i>=0;i--){
				self.queuedTransferTable.removeRow(rows[i]);
				self.taskList.splice(rows[i], 1);
			}
			self._nextTask();
		});
	}else{
		if (rows[0]==0){ // but not running
			if (self.taskList[0].type=='GET'){
				if (self.currentWriteStream && self.currentWriteStream.writable){
					self.currentWriteStream.destroy();
					self.currentWriteStream = false;
				}
				var fn = self.taskList[0].dest;
				if (fs.existsSync(fn)){
					fs.unlinkSync(fn);
				}
			}
		}
		for (var i=rows.length-1;i>=0;i--){
			self.queuedTransferTable.removeRow(rows[i]);
			self.taskList.splice(rows[i], 1);
		}
	}
}

/**
 * Starts processing the transfer list
 *
 * @this {TransferManager}
 */
TransferManager.prototype.start = function(){
	var self = this;

	if (self.taskList.length && !self.running){
		self.running = true;
		self._nextTask();
	}
}

/**
 * Stops processing the transfer list, NOT the currently running FTP transfer ??
 *
 * @this {TransferManager}
 */
TransferManager.prototype.stop = function(){
	var self = this;
	if (self.running){

		if (self.uploadTimer){
			clearInterval(self.uploadTimer);
			self.uploadTimer = undefined;
		}
		if (self.tmpClient.connected){
			self.tmpClient.abort(true, function(){
				//console.log('tmpClient aborted');
			});
		}

		self.client.abort(true, function(){
			self.running = false;
			self.connected = false;
			self.currentSite = undefined; // string (prop name of global.sites)
			// reset status in table to 0%?
		});
	}
}

/**
 * Returns the container widget (a QTabWidget in this case)
 *
 * @this {TransferManager}
 */
TransferManager.prototype.container = function(){
	var self = this;

	return self.tabWidget;
}

/**
 *
 *
 * @this {TransferManager}
 */
TransferManager.prototype.saveTasks = function(){
	var self = this;

	fs.writeFileSync('./config/tasks.json', JSON.stringify(self.taskList, null, '  '), 'binary');
}

/**
 *
 *
 * @this {TransferManager}
 */
TransferManager.prototype.loadTasks = function(){
	var self = this;

	try{
		var taskList = JSON.parse(fs.readFileSync('./config/tasks.json', 'binary'));
		fs.unlinkSync('./config/tasks.json');
		self.addTransfers(taskList);
	}catch(e){}
}

return TransferManager;
})();// END CLASS
